import numpy as np
import matplotlib.pyplot as plt

def band_ratio(data, num, den):
    """
    Calculate a band ratio to map broad absorbion features (e.g. iron).

    Args:
        data: HyData instance to calculate a band ratio for.
        num: the numerator of the band ratio. Integers are treated as indices, floats as wavelengths.

                If a tuple is passed then values between band1[0] and band1[1] will be averaged before computing the
               band ratio. Lists of tuples/floats or indices will be summed (for calculating ratios of the form (a+b)/c).

        den: the denominator of the band ratio. Values are treated like in band1.
    Returns:
        a new HyData instance containing the band ratio.

    """
    if not isinstance(num, list):
        num = [num]
    if not isinstance(den, list):
        den = [den]

    # calculate numerator and denominator
    numdata = np.zeros(data.data.shape[:-1])
    dendata = np.zeros(data.data.shape[:-1])
    names = []
    for D, exp in zip([numdata, dendata], [num, den]):
        name = None
        for n in exp:
            if isinstance(n, tuple):
                # calculate addition or subtraction
                sign = 1
                if n[0] < 0 or n[1] < 0:
                    sign = -1
                    n = (abs(n[0]), abs(n[1]))

                assert n[0] < n[1], "Error - invalid slice %.1f:%.1f" % n
                idx = slice(data.get_band_index(n[0]),
                            data.get_band_index(n[1]), 1)
                if name is None:
                    name = "%.1f:%.1f" % n
                else:
                    if sign > 0:
                        name += " + %.1f:%.1f" % n
                    else:
                        name += " - %.1f:%.1f" % n
            else:
                sign = 1
                if isinstance(n,str):
                    if n[0] == '-':
                        n = n[1:]
                        sign = -1
                elif n < 0:
                    sign = -1
                    n = abs(n)

                idx = data.get_band_index(n)
                if name is None:
                    if isinstance(n,str):
                        name = "%s" % n
                    else:
                        name = "%.1f" % n
                else:
                    if sign > 0:
                        if isinstance(n, str):
                            name = "+%s" % n
                        else:
                            name += "+%.1f" % n
                    else:
                        if isinstance(n, str):
                            name = "-%s" % n
                        else:
                            name += "-%.1f" % n

            # accumulate
            slc = data.data[..., idx]
            if isinstance(idx, slice):
                slc = np.nanmean(slc, axis=-1)
            D += sign * slc

        # store name
        names.append(name)

    # calculate output
    out = data.copy(data=False)
    out.header.drop_all_bands()  # drop band specific attributes
    out.data = np.nan_to_num( (numdata / dendata)[..., None], posinf=0, neginf=0 ) # replace infs, nans, etc. with 0

    # set band name
    out.set_band_names(['(%s) / (%s)' % (names[0], names[1])])
    out.push_to_header()

    # return HyData instance
    return out


def NDVI(data):
    """
    Calculate NDVI.

    Args:
        data: the HyData instance to calculate a NDVI for.
    Returns:
        a new HyData instance containing the band ratio.
    """

    idxNIR = data.get_band_index(800.0)
    idxRed = data.get_band_index(670.0)

    # create output HyData instance
    out = data.copy(data=False)
    out.header.drop_all_bands()  # drop band specific attributes

    # calculate NDVI
    with np.errstate(all='ignore'):  # ignore div0 errors
        out.data = ((data.data[..., idxNIR] - data.data[..., idxRed]) / (
                data.data[..., idxNIR] + data.data[..., idxRed]))[..., None]

    # set band name
    out.set_band_names(['NDVI'])
    out.push_to_header()

    return out

def SKY(data):
    """
    Calculate ratio between blue (479.89 nm and featureless swir (1688.64 nm) to identify sky pixels and (sometimes)
    very distance pixels due to blue scattering.

    Args:
        data: the HyData instance to calculate a NDVI for.
    Returns:
        a new HyData instance containing the SKY band ratio.
    """

    return band_ratio(data, 479.89, 1688.64)

def SHADE(data):
    """
    Calculate ratio between blue (479.89 nm and near infrared (1688.64 nm) to identify shadow pixels (at least in theory).

    Generally this should work well, except for objects that are very blue. Luckily blue things are not common in Geology.

    Args:
        data: the HyData instance to calculate SHADE for.
    Returns:
        a new HyData instance containing the SHADE band ratio.
    """

    return band_ratio(data, 480.0, 800.0)




"""
A collection of functions for performing illumination corrections to derive reflectance estimates
from radiance data.
"""

import datetime
import numpy as np
import pytz
from scipy import stats
import datetime
from hylite.correct import get_hull_corrected
import matplotlib.pyplot as plt
import numpy as np

#####################
## Utility functions
######################
from hylite.correct import Panel

from .occlusion import *
from .reflection import *
from .path import *

def sph2cart(az, el, r=1.0):
    """
    Convert spherical coordiantes to cartesian ones.
    """

    az = np.deg2rad(az)
    el = np.deg2rad(el)
    return np.array( [
        np.sin(az) * np.cos(el),
        np.cos(az) * np.cos(el),
        -np.sin(el) ]) * r

def cart2sph(x, y, z):
    """
    Convert cartesian coordinates to spherical trend, plunge and radius.
    """

    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    az = np.arctan2(x, y)
    el = np.arcsin(z / r)

    while az < 0:
        az += (2*np.pi)
    return np.array([np.rad2deg(az), -np.rad2deg(el), r])

def estimate_sun_vec(lat, lon, time):
    """
    Calculate the sun illumination vector at the specified position and time.

    Args:
        lat: the latitude of the position to calculate the sun vector at (in decimal degrees).
        lon: the longitude of the position to calculate the sun vector at (in decimal degrees).
        time: the time the dataset was acquired. This, and the position defined by the "pos"
              argument will be used to calculate the sun direction. Should be an instance of datetime.datetime,
              or a tuple containing (timestring, formatstring, pytz timezone).
              E.g. time = ("19/04/2019 12:28","%d/%m/%Y %H:%M", 'Europe/Madrid')

    Returns:
        A tuple containing:

         - sunvec = the sun illumination direction (i.e. from the sun to the observer) in cartesian coords
         - azimuth = the azimuth of the sun (bearing towards sun)
         - elevation = the elevation of the sun (angle above horizon)

    """

    # get time
    if isinstance(time, tuple):  # parse time from strings
        tz = time[2]
        time = datetime.datetime.strptime(time[0], time[1])
        tz = pytz.timezone(tz)
        time = tz.localize(time)

    # time = time.astimezone(pytz.utc) #convert to UTC

    assert isinstance(time, datetime.datetime), "Error - time must be a datetime.datetime instance"
    assert not time.tzinfo is None, "Error - time zone must be specified (e.g. using tz='timezone')."

    # calculate illumination vector from time/position
    import astral.sun
    pos = astral.Observer(lat, lon, 0)
    azimuth = astral.sun.azimuth(pos, time)
    elevation = astral.sun.elevation(pos, time)

    sunvec = sph2cart(azimuth + 180, elevation)  # n.b. +180 flips direction from vector point at sun to vector
    # pointing away from the sun.

    return sunvec, azimuth, elevation

def estimate_skyview( normal ):
    """
    Return a crude estimate of sky-view factor by assuming geometry is an inclined plane (and that the sky is invisible
    below the horizon level. Useful for e.g. estimating the skyview factor for calibration panels based on their orientation.

    Args:
        normal: the surface or panel normal vector as a (3,) numpy array or (...,3) image or cloud.
    Returns:
        the skyview factor (0 - 1).
    """
    return ( np.pi - np.arccos( np.dot( normal, np.array([0,0,1]) ) ) ) / np.pi # sky view factor for inclined plane

def _regress(x, y, split=True, clip=(10, 90)):
    """ Loop through all bands in specified arrays and fit a line. Used for statistical adjustments (e.g. c-factor)"""

    assert x.shape == y.shape, "Error - x and y have incompatible shapes (%s != %s)." % (x.shape, y.shape)

    if not split:  # easy!
        x = x.ravel()
        y = y.ravel()

        # calculate percentile clip
        xmn, xmx = np.nanpercentile(x, clip)
        ymn, ymx = np.nanpercentile(y, clip)

        # calculate valid mask
        mask = np.isfinite(x) & np.isfinite(y) & (x > xmn) & (y > ymn) & (x < xmx) & (y < ymx)
        if not mask.any():  # no data to regress... return boring line.
            i = 0.0
            s = 0.0
            r = np.array([[np.nan]])
        else:
            (i, s), r = np.polynomial.polynomial.polyfit(x[mask], y[mask], 1, full=True)

            # plt.plot(x,y,alpha=0.2)
            # sx = np.nanmin(x[mask])
            # ex = np.nanmax(x[mask])
            # plt.plot( [sx,ex], [s*sx+i, s*ex+i])
            # plt.show()
            # assert False
        return s, i, np.sqrt((r[0][0] / np.sum(mask)))
    else:
        out = []

        # loop through bands
        for b in range(x.shape[1]):
            out.append(_regress(x[:, b], y[:, b], split=False))  # do regression

        # return output
        out = np.array(out)
        return out[:, 1], out[:, 0], out[:, 2]


#############################################################################
## Useful functions for building / doing illumination corrections
#############################################################################

def UAC(data, band_range=(0, -1), thresh=98, vb=True):
    """
    Apply a universal absorption correction. This is a generalised version of the algorithm presented by
    Lorenz et al., 2018 to identify and remove absorption features that are present in all of the spectra
    in a dataset, as these generally relate to residual atmospheric or sensor effects.

    Reference:
    https://doi.org/10.3390/rs10020176

    Args:
        image: a hyperspectral image to correct
        band_range: a range of bands to do this over. Default is (0,-1), which applies the correction to all bands.
        thresh: the percentile to apply when identifying the smallest absorbtion in any range based on hull corrected
                spectra. Lower values will remove more absorption (potentially including features of interest).
        vb: True if a progress bar should be created during hull correction steps.

    Returns:
        a HyData instance containing the corrected spectra.
    """
    # subset dataset
    out = data.export_bands(band_range)
    nanmask = np.logical_not(np.isfinite(out.data))
    out.data[nanmask] = 0  # replace nans with 0

    # do hull correction
    hc = get_hull_corrected(out, vb=vb)

    # identify maximum hull
    mx = np.nanpercentile(hc.X(), thresh, axis=0)

    # apply adjustment and return
    if out.is_image():
        out.data /= mx[None, None, :]
    else:
        out.data /= mx[None, :]
    out.data[nanmask] = np.nan  # add nans back in

    return out


def estimate_illu(image, panel, pilf, pskv, ilf, skv=0.6, oc=None, thresh=0.01, clip=None):
    """
    Estimate the sky and sunlight spectra by comparing shaded vs non-shaded regions in a dataset and assuming
    they have the same median reflectance spectra. See Thiele et al., 2021 for more details:
    https://doi.org/10.1109/TGRS.2021.3098725.

    Args:
        image: the hyperspectral image to extract spectra from.
        panel: a Panel instance containing calibration spectra from a fully illuminated calibration panel.
        pilf: a measured or estimated lambert illumination factor for the panel (from 0 to 1). If a value of
              0 is used then it is assumed that the panel is completely shaded.
        pskv: a measured or estimated skyview factor for the panel (from 0.1 to 1).
        ilf: direct illumination factors representing the fraction of downwelling light reflected towards the
             sensor, as estimated using e.g., hylite.correct.illumination.reflection.calcLambert(...) or
             hylite.correct.illumination.reflection.calcOrenNayar(..).
        skv: a(width,height) array of skyview factors or a float containg the average sky view factor (0 to 1)
             for the scene. Default is 0.6.
        oc: a (width, height) array of occlusion factors computed using
            e.g., hylite.correct.illumination.occlusion.calcBandRatioOcc, or None (default; for no cast shadows).
        thresh: the threshold for direct illumination factor (skv * (1-oc)) at which a pixel is considered entirely
                lit by ambient (sky) light. Default is 0.01.
        clip: a subset of the image to use for estimation (to ensure our assumption of equal median reflectance between
              sun and shade pixels is met). Default is None, but a clipping rectangle can be passed as (xmin,xmax,ymin,ymax).
    """

    assert pskv > 0.1, "Error - panel skyview factor must be > 0.1, not %.2f" % (pskv)
    assert pilf < 1 or pskv < 1, "Error - panel illumination factors must both be < 1, not %.2f, %.2f" % (pskv, pilf)

    assert ilf.shape[0] == image.data.shape[0], "Error - incompatible width between image and ilf array."
    assert ilf.shape[1] == image.data.shape[1], "Error - incompatible width between image and ilf array."

    # extract radiance samples
    if clip is None:
        clip = (0, -1, 0, -1)
    r = image.data[clip[0]:clip[1], clip[2]:clip[3], :].reshape(-1, image.band_count())  # radiance measurements
    i = ilf.squeeze()[clip[0]:clip[1], clip[2]:clip[3]].ravel()  # illumination factors
    if isinstance(skv, float):
        a = np.full(i.shape[0], skv)
    else:
        assert isinstance(skv, np.ndarray), "Error - skv must be a numpy array not a %s" % type(skv)
        assert skv.shape[0] == image.data.shape[0], "Error - incompatible width between image and skv array."
        assert skv.shape[1] == image.data.shape[1], "Error - incompatible width between image and skv array."
        a = skv.squeeze()[clip[0]:clip[1], clip[2]:clip[3]].ravel()

    # apply occlusion mask
    if oc is not None:
        assert isinstance(oc, np.ndarray), "Error - oc must be a numpy array not a %s" % type(oc)
        assert oc.shape[0] == image.data.shape[0], "Error - incompatible width between image and oc array."
        assert oc.shape[1] == image.data.shape[1], "Error - incompatible width between image and oc array."
        o = oc.squeeze()[clip[0]:clip[1], clip[2]:clip[3]].ravel()  # extract occlusion data
        i = i * (1 - 0)  # apply it to our direct reflectance fractions

    # extract shaded pixel mask
    shadepix = (i <= thresh)

    # compute normalised difference between sun and shade
    delta = np.nanmedian(a[shadepix, None] / r[shadepix, :], axis=0) - np.nanmedian(a[..., None] / r[..., :], axis=0)
    delta /= np.nanmedian(i[..., None] / r[..., :], axis=0)

    # derive skylight and sunlight spectra
    rfrac = (panel.get_mean_radiance() / panel.get_reflectance())
    skyest = rfrac / (pskv + delta * pilf)
    sunest = skyest * delta

    return sunest, skyest
#############################################################################
## Define generic illumination model
#############################################################################

class IlluModel(object):
    """
    Combine source, reflection, occlusion and transmittance information to simulate
    the illumination within a scene.

    Fundamentally, an illumination model represents one or all parts of the following
    equation:

    r = (aS + oI)R + P

    Where,

     - r = the measured pixel radiance.
     - a = the skyview factor associated with this pixel.
     - S = the downwelling skylight spectra (assumed to be constant across the scene).
     - o = the illumination factor for this pixel; will combine occlusion and e.g. lambert or oren-nayar reflection.
     - I = the incident sunlight spectra in the scene.
     - R = the material reflectance (0 - 1) of the pixel. Typically this is what we are trying to solve for.
     - P = path radiance effects that add or remove light due to interactions with air or particles between the
           sensor and the target.
    """

    def __init__(self, I, S=0, P=0, skv=0.6, rf=1.0, oc=0):
        """
        Create an illumination model. All arguments should be (...,1) shape arrays.

        Args:
            I: the incident sunlight spectra. This must be provided.
            P: the path radiance spectra. Defaults to 0 if unknown (i.e. ignore path radiance). If this
               is a 1-D array then it is assumed to be constant across the scene. If it is an (...,1) shaped
               array then it will be applied separately per-pixel or per-point, to account for e.g. variations
               in distance to target.
            S: the downwelling skylight spectra. Defaults to 0 if unknown (i.e. ignore skylight).
            skv: skyview factors. Will be ignored unless a skylight spectra is defined. Default is 0.6
                     (if a skylight spectra is defined but skyview factors are unknown).
            rf: reflectance factors as determined by e.g. a lambertian or oren-nayar reflectance model. Default
                is a perfect reflection (1.0).
            o: occlusion factors that reduce the amount of incident light recieved across the scene due to e.g.
               shadows. Reflection factors specified by the rf argument are multiplied by (1 - oc). Default is 0
               (no occlusion).

        Returns:
            an IlluModel instance.
        """
        # store data
        self.I = I  # n.b. this copies arrays (for safety), and also wraps constants in a numpy object.
        self.P = P
        self.S = S
        self.skv = skv
        self.rf = rf

        if oc is None:
            self.oc = 0 # no occlusions
        else:
            self.oc = oc

        # initialise fitted attributes
        self.rboost = np.array(0)  # radiance boost [ c-factor adjustment in y ]
        self.iboost = np.array(0)  # illumination boost [ c-factor adjustment in x ]

        # make all attributes numpy arrays
        for attr in ['I', 'P', 'S', 'skv', 'rf', 'oc']:
            val = getattr(self, attr)
            if isinstance(val, hylite.HyData):  # hydata instance
                setattr(self, attr, np.array(val.data).squeeze().astype(np.float32))  # copy out data array
            else:
                setattr(self, attr, np.array(val).squeeze().astype(np.float32))

    def getRadiance(self, R=1.0):
        """
        Evaluate the measured radiance for a known reflectance, based on the equation:

        r = (aS + oI)R + P

        Args:
            R: the reflectance(s) to evaluate. If a scalar value is passed (must be between 0 and 1) then
               a single combined illumination spectra will be returned. If a (...,1) array of values are
               provided then a (...,n) array containing the illumination spectra at each point or pixel will
               be returned.
        Returns:
            an array containing the illumination spectra in its last axis.
        """
        if isinstance(R, hylite.HyData):
            R = R.data.squeeze()
        else:
            R = np.array(R).squeeze()
        if len(R.shape) == 0 or R.shape[-1] == 1:  # R is constant in the spectral direction
            return (self.skv[..., None] * self.S + ((1 - self.oc) * self.rf)[..., None]) * R[
                ..., None] * self.I + self.P
        else:
            return (self.skv[..., None] * self.S + ((1 - self.oc) * self.rf)[..., None]) * R * self.I + self.P

    def getReflectance(self, r, strict=True):
        """
        Evaluate the reflectance from measured radiance, based on the equation:

        R = (r - P) / (aS + oI)

        Args:
            r: a (..., n) array containing the measured radiance values.
            strict: True if reflectance values should be clipped to the physically plausible range (0 - 1). Default is True.
        Returns:
            an array containing the reflectance spectra in its last axis.
        """
        if isinstance(r, hylite.HyData):
            self.r = r.data.squeeze()
        else:
            self.r = np.array(r).squeeze()

        self.R = (self.r - self.P + self.rboost) / (
                self.skv[..., None] * self.S + ((1 - self.oc) * self.rf)[..., None] * self.I + self.iboost)
        if strict:
            self.R = np.clip(self.R, 0, 1)

        return self.R

    def fit(self, radiance, shift='x'):
        """
        Compute c-factor offsets. This assumes a linear relationship between illumination and measured radiance, and
        increases/decreases either (1) the illumination component [y-shift] or (2) the measured radiance [x-shift] to
        force the regression line to pass through (0,0). Reflectance outliers are then detected and can be masked or
        corrected.

        Args:
            radiance: radiance data to fit to. Must have the same shape as self.data.
            shift: apply the correction in the y-direction (adjust measured radiance to simulate the influence of path
                   radiance) or in the x-direction (adjust modelled illumination to account for unknown light source). Default
                   is 'x' (this is the typical c-factor correction).
        """

        # compute estimated radiance
        x = self.getRadiance()
        x = x.reshape((-1, x.shape[-1]))

        # get measured radiance
        y = radiance.X()

        # remove any mischevous negative radiances... (these are noise)
        x[x < 0] = np.nan

        # compute slope and intercepts of regression
        s, i, r = _regress(x, y)  # fit linear regressions
        if 'x' in shift:  # typical c-factor adjustment (add/subtract illumination)
            self.iboost = np.array(i) / np.array(s)  # calculate cfacor (x-intercept)
            mn = np.nanmin(x, axis=0)  # calculate floor cfactor to avoid negative values
            self.iboost[self.iboost < mn] = mn[self.iboost < mn]  # apply floor and store
        elif 'y' in shift:  # alternative adjustment ("r-factor"); add/subtract irradiance.
            self.rboost = -np.array(s)  # calculate alternative c-factor (y-intercept) and store
        else:
            assert False, 'Error - %s should be either "x" or "y"' % shift

    def plot_fit(self, radiance, bands=None, n=100, nb=5, **kwds):
        """
        Plot the relationship between illumination and measured radiance.

        Args:
            radiance: the radiance data (HyImage or HyCloud) to compare too. Shape must match internal self.data array.
            bands: the band (integer or float), band range (tuple) or bands (list) to include on the regression plot. Default
                   is None (use all bands).
            n: plot every nth point (only) to speed up plotting. Default is 100. This value does not affect the regressions.
            nb: only calculate / plot every nb'th band if bands is a (min,max) tuple. Default is 5.
        *Keywords*:
            keywords are passed to plt.scatter(...).
        """

        # get data
        x = self.getRadiance()
        x = x.reshape((-1, x.shape[-1]))
        y = radiance.X()

        # get relevant bands
        if bands is None:
            bands = (0, -1)
        if isinstance(bands, float) or isinstance(bands, int):
            idx = radiance.get_band_index(bands)
            x = x[:, idx][:, None]
            y = y[:, idx][:, None]
            w = radiance.get_wavelengths()[idx]
        elif isinstance(bands, tuple) and len(bands) == 2:
            mn, mx = [radiance.get_band_index(b) for b in bands]
            idx = np.array(range(radiance.band_count()))[mn:mx]  # band indices
            x = x[:, idx][:, ::nb]
            y = y[:, idx][:, ::nb]
            w = radiance.get_wavelengths()[idx][::nb]
        elif isinstance(bands, list) or isinstance(bands, tuple):
            idx = np.array([radiance.get_band_index(b) for b in bands])
            nb = 1
            x = x[:, idx][::nb]
            y = y[:, idx][::nb]
            w = radiance.get_wavelengths()[idx][::nb]
        else:
            assert False, "Error - %s is an unknown band type. Should be int, float, list or tuple." % type(bands)

        # remove any mischevous negative radiances... (these are noise)
        x[x < 0] = np.nan

        # build plot
        if len( self.iboost.shape ) == 0 and len( self.rboost.shape ) == 0:  # no adjustment applied
            fig, ax = plt.subplots(2, 1, figsize=(10, 10))  # only two axes needed if no shift applied
            ax = [ax[0], ax[0], ax[1]]
        else:
            fig, ax = plt.subplots(3, 1, figsize=(10, 10))

        ###################
        # (a) plot points
        ###################
        cmap = plt.get_cmap('rainbow')
        kwds['s'] = kwds.get('s', 3.0)
        kwds['alpha'] = kwds.get('alpha', 0.1)
        for b in range(x.shape[-1]):
            kwds['color'] = cmap((w[b] - np.min(w)) / np.ptp(w))
            ax[0].scatter(x[::n, b], y[::n, b], **kwds)
        ax[0].set_xlabel("Modelled radiance")
        ax[0].set_ylabel("Measured radiance")
        ax[0].set_title("a. Radiance")

        ###################
        # (b) plot c-factor
        ###################
        if len( self.iboost.shape ) != 0:
            ax[1].plot(radiance.get_wavelengths()[idx], self.iboost[idx])
            ax[1].axhline(0, color='k', lw=2)
            ax[1].set_ylabel("Illumination boost")
            ax[1].set_xlabel("Wavelength (nm)")
            ax[1].set_title("b. Illumination boost adjustment (c-factor adjustment)")

            # apply adjustment for plotting
            x += self.iboost[idx][::nb]

        elif len( self.rboost.shape ) != 0:  # n.b. cfac and rfac should never both be set
            ax[1].plot(radiance.get_wavelengths()[idx], self.rboost[idx])
            ax[1].axhline(0, color='k', lw=2)
            ax[1].set_ylabel("Irradiance boost")
            ax[1].set_xlabel("Wavelength (nm)")
            ax[1].set_title("b. Radiance boost adjustment (alternate c-factor adjustment)")

            # apply adjustment for plotting
            y += self.rboost[idx][::nb]

        #############################################
        # (b) plot reflectance vs at target radiance
        #############################################
        y = y / x  # convert to reflectance

        cmap = plt.get_cmap('rainbow')
        kwds['s'] = kwds.get('s', 3.0)
        kwds['alpha'] = kwds.get('alpha', 0.1)
        for b in range(x.shape[-1]):
            kwds['color'] = cmap((w[b] - np.min(w)) / np.ptp(w))
            ax[2].scatter(x[::n, b], y[::n, b], **kwds)

        p5, p10, p50, p90, p95 = np.nanpercentile(y, (5, 10, 50, 90, 95))
        ax[2].axhline(p10, color='k', ls='--')
        ax[2].axhline(p90, color='k', ls='--')
        ax[2].axhline(p50, color='k', lw=2)

        ax[2].set_ylim(p5, p95)
        ax[2].set_ylabel("Reflectance")
        ax[2].set_xlabel("Modelled Radiance")
        ax[2].set_title("c. Reflectance")

        # add a wavelength colorbar
        sc = ax[0].scatter([np.nan, np.nan], [np.nan, np.nan], c=[np.min(w), np.max(w)],
                           vmin=np.min(w), vmax=np.max(w), cmap='rainbow')  # build colormap
        cbar = fig.colorbar(sc, orientation='horizontal', shrink=0.5, pad=0.25)  # plot
        cbar.set_label('Wavelength (nm)')

        fig.tight_layout()
        return fig, ax


class ELC(object):
    """
    Class that gathers one or more Panels and computes calculates an empirical line correction. This does not
    adequately describe or correct for scene illumination, but can be useful as a quick correction step.
    """

    def __init__(self, panels):

        """
        Constructor that takes a list of Panel objects (one for each target used for the correction) and computes
        an empirical line correction.

        Args:
             panels: a list of Panel objects defining the reflectance and radiance of each panel in the scene.
        """

        if not isinstance(panels, list):
            panels = [panels]

        self.wav = np.array(panels[0].get_wavelengths())
        for p in panels:
            assert isinstance(p, Panel), "Error - ELC panels must be instances of hylite.correct.Panel"
            assert (self.wav == np.array(
                p.get_wavelengths())).all(), 'Error - ELC panels must cover the same wavelengths'

        # compute ELC
        self.slope = np.zeros(self.wav.shape)
        self.intercept = np.zeros(self.wav.shape)
        if len(panels) == 1:  # only one panel - assume intercept = 0
            self.slope = panels[0].get_mean_radiance() / panels[0].get_reflectance()
        else:
            # calculate regression for each band
            for b, w in enumerate(self.wav):
                _y = np.array([p.get_mean_radiance()[b] for p in panels])
                _x = np.array([p.get_reflectance()[b] for p in panels])
                self.slope[b], self.intercept[b], _, _, _ = stats.linregress(_x, _y)

    def get_wavelengths(self):
        """
        Get the wavelengths for which this ELC has been calculated.
        """
        return self.wav

    def get_bad_bands(self, **kwds):

        """
        Find bands in which signal-noise ratios are amplified above a threshold (due to large correction slope).

        Args:
            **kwds: Keywords can include:

                - thresh = the threshold slope. Defaults to the 85th percentile.

        Returns:
            a boolean numpy array containing True for bad bands and False otherwise.
        """

        thresh = kwds.get("thresh", np.nanpercentile(self.slope, 85))
        return self.slope > thresh

    def apply(self, data, **kwds):

        """
        Apply this empirical line calibration to the specified image.

        Args:
            data: a HyData instance to correct
            **kwds: Keywords can include:

                 - thresh = the threshold slope. Defaults to the 90th percentile.

        Returns:
            a mask containing true where the corrected values are considered reasonable - see get_bad_bands(...) for more
            details. Note that this returns the np.logical_not( self.get_bad_bands(...) ).
        """

        assert data.band_count() == len(self.slope), "Error - data has %d bands but ELC has %d" % (
        data.band_count(), len(self.slope))
        data.data -= self.intercept # subtract path radiance
        data.data /= self.slope # remove illumination source

        return np.logical_not(self.get_bad_bands(**kwds))

    def quick_plot(self, ax=None, **kwds):

        """
        Plots the correction factors (slope and intercept) computed for this ELC.

        Args:
            ax: the axes to plot on. If None (default) then a new axes is created.
            **kwds: Keywords can include:

                 - figsize = a figsize for the figure to create (if ax is None).
                 - thresh = the threshold to separate good vs bad correction values (see get_bad_bands(...)). Default is the
                            85th percentile of slope values.
        Returns:
            fig, ax = the figure and axes objects containing the plot.

        """

        if ax is None:
            fig, ax = plt.subplots(figsize=kwds.pop('figsize',(15, 10)))

        # plot slope
        _x = self.get_wavelengths()
        _y1 = self.slope
        _y2 = [kwds.get("thresh", np.nanpercentile(self.slope, 85))] * len(_y1)
        ax.plot(_x, _y1, color='k', lw=1)
        ax.plot(_x, _y2, color='gray', lw=2)
        ax.fill_between(_x, _y1, [0] * len(_x), where=_y1 > _y2, facecolor='red', interpolate=True, alpha=0.3)
        ax.fill_between(_x, _y1, [0] * len(_x), where=_y1 < _y2, facecolor='green', interpolate=True, alpha=0.3)
        ax.set_xlabel("Wavelength (nm)")
        ax.set_ylabel("ELC slope")
        if not (self.intercept == 0).all():
            ax2 = ax.twinx()
            ax2.plot(_x, self.intercept, color='b')
            ax2.set_ylabel("ELC intercept")
            ax2.spines['right'].set_color('blue')
            ax2.yaxis.label.set_color('blue')
            ax2.tick_params(axis='y', colors='blue')

        return ax.get_figure(), ax


"""
Load or import spectral libraries.
"""

import os
import numpy as np
import glob
import hylite
from hylite import HyLibrary
from hylite.io import makeDirs
from time import gmtime, strftime
from hylite.io.images import loadWithGDAL, saveWithGDAL, loadWithSPy, loadWithSPy

# noinspection PyUnusedLocal
def _read_sed_file(path):

    """
    Read a (single) reference spectra from a .sed file.

    Args:
        path to the file.

    Returns:
        A tuple containing:

             - wav = a list of the wavelengths in the sed file
             - refl = a list of corresponding reflectances
             - name = a name for this sample (the file name without extension)
             - meta = a dictionary containing additional metadata in the header to the .sed file
     """

    with open(path, 'r') as f:
        # read header into metadata dict
        meta = {}
        l = f.readline()
        while not 'data:' in l.lower():
            splt = l[:-1].split(':')
            if len(splt) == 1:
                meta[splt[0]] = ""  # just store flag
            else:
                meta[splt[0]] = ''.join(splt[1:]).strip()
            l = f.readline()

        # get dataset name
        name = os.path.splitext(os.path.basename(path))[0]

        # read colum names
        # n.b. we ignore these for now. We could parse them by splitting by spaces EXCLUDING spaces
        # that are preceded by a '.' (i.e. split on ' ' but not '. ').
        col_names = f.readline()

        # read data lines
        wav = []
        refl = []
        l = f.readline()
        while l:
            # n.b. lines will be something like
            # 344.2	1.210000E+001	1.030000E+000	  7.815
            data = np.fromstring(l, sep='\t')
            wav.append(data[0])  # wavelength always first?
            refl.append(data[-1])  # reflectance always last?
            l = f.readline()
        # return
        return wav, refl, name, meta

def loadLibrarySED(path):
    """
    Load a spectral library stored as a folder of .sed files.

    Args:
        directory: the directory containing the .sed files. Filenames will be used as sample names.
    """

    assert os.path.isdir(path), "Error - must specify a directory, not a file."
    import utm

    # get sed files
    files = glob.glob(path + '/*.sed')
    assert len(files) > 0, "Error - no .sed files found in directory (%s)" % path

    # load sed files
    wav = None
    meta = None
    data = []
    names = []
    pos = []
    for f in files:
        _w, _r, _n, _m = _read_sed_file(f)  # read sed file

        # set or check wavelengths
        if wav is None:
            wav = _w
            meta = _m
        elif wav != _w:
            print("Error - .sed files have incompatable wavelength information")
            print("   Most recently read file has %d wavelengths. ", len(_w))
            print("   First file has %d wavelenghts.: ", len(wav))
            assert False

        data.append(_r)
        names.append(_n)

        if 'Latitude' in _m and 'Longitude' in _m and 'Altitude' in _m:
            try:
                _x = float(_m['Longitude'])
                _y = float(_m['Latitude'])
                _z = float(_m['Altitude'])

                # convert pos to UTM
                _x, _y, zn, zl = utm.from_latlon(_x, _y)
                meta['UTM Zone'] = '%d%s' % (zn, zl)
                pos.append([_x, _y, _z])
            except:
                pass  # not a float...

    # create spectral library
    library = HyLibrary(np.array(data), names, wav=np.array(wav))

    # add metadata to header
    ignore = ['comment', 'version', 'file name', 'columns [4]']
    for key, val in meta.items():
        if key.lower() in ignore:
            continue
        library.header[key] = val

    return library

def loadLibraryTSG(path):
    """
    Load a spectral library of TSG (The Spectral GeologistTM, e.g. HyLogger data) stored as a folder of .csv files.

    Args:
        directory: the directory containing the .sed files. Filenames will be used as sample names.
    """

    assert ".csv" in path, "Error - path must specify a csv file."

    # load TSG files
    meta = []
    refl = []
    pos = []

    # get dataset name
    name = os.path.splitext(os.path.basename(path))[0]

    with open(path, 'r') as f:
        # read wav
        wav_str = f.readline()
        wav = np.fromstring(wav_str.strip("Wavelength,"), sep=",")

        # read data lines
        l = f.readline()
        while l:
            # n.b. lines will be something like
            # NC3_0001_0009__T=1_L=1_P=9_D=6.0000003_X=559	0.0798395	0.0769894	0.0747408	0.0729613
            meta.append(l.split(",", 1)[0])
            refl.append(np.fromstring(l.split(",", 1)[1], sep=','))
            pos.append(float(((l.split(",", 1)[0]).split('D=' ,1)[1]).split('_X=' ,1)[0]))
            l = f.readline()
    names = [name +"_ " +str(x) for x in pos]

    # create spectral library
    library = HyLibrary(np.array(refl), names, wav=np.array(wav))
    library.header["depth"] = pos

    return library

def saveLibraryCSV(path, library):
    """
    Save a spectral library to a csv file with wavelengths as columns and sample reflectance as rows
    """

    # make directories if need be
    makeDirs(path)

    with open(path, 'w') as f:
        f.write( "name," + str(list(library.get_wavelengths()))[1:-1] +'\n' )
        for i,n in enumerate(library.get_sample_names()):
            f.write( str(n) + ", " + str(list( library.data[i,0,:] ))[1:-1] + '\n' )
    f.close()

def loadLibraryCSV(path):
    with open(path, 'r') as f:
        header = f.readline()
        wav = np.array( header.split(",")[1:], dtype=float)

        # parse spectra
        names = []
        refl = []
        l = f.readline()
        while l:
            l = l.split(",")
            names.append( l[0] )
            refl.append( np.array(l[1:],dtype=float) )
            l = f.readline()
    return HyLibrary( np.array(refl), names, wav=wav )


def loadLibraryTXT(path):
    """
    Load an ENVI text format library. This should have the following structure:

            ENVI ASCII Plot File
            Column 1: Wavelength
            Column 2: Sample1
            Column 3: Sample2
            C10 C20 C30
            C11 C21 C31
            C12 C22 C32
            ... ... ..

    """

    with open(path, 'r') as f:
        # check header line
        l0 = f.readline()
        assert 'envi ascii' in l0.lower(), "Error - Provided file is not an ENVI TXT library: %s" % path

        # read sample definitions
        ll = f.readline().lower()
        names = {}
        wav = -1
        i = 0
        while 'column' in ll:

            # split and get column number
            if ':' in ll:
                cname = ll.strip().split(':')[1]

                if 'wav' in cname:  # this column is wavelengths
                    wav = i
                else:  # this column is a spectra
                    names[i] = cname
                i += 1

            # read next line
            ll = f.readline().lower()

        # read data block
        data = []
        while ll:
            if ',' in ll:  # comma separated
                data.append(np.fromstring(ll, sep=','))
            else:  # space separated
                data.append(np.fromstring(ll, sep=' '))
            ll = f.readline()
        data = np.array(data).T
        f.close()  # close file

        # build hylibrary
        if wav > -1:
            wav = data[wav, :]
        else:
            wav = np.arange(data.shape[-1])
        lib = HyLibrary(data[list(names.keys()), :], lab=list(names.values()), wav=wav)
        return lib


import glob


def loadLibraryDIR(path, wav=None):
    """
    Load a spectral library from a directory of ENVI text files, with the following structure:

     - path
         - mineralA
             - spectraA.txt
             - spectraB.txt
             - spectraC.txt
         - mineralB
             -spectraA.txt
             -spectraB.txt
        etc.

    Data in this format can be downloaded using iSpec: https://www.samthiele.science/app/iSpec/index.html.


    Args:
        path: the directory path to search for ENVI spectra.
        wav: an array of wavelengths to resample the spectra onto. This is required as the txt files often
             have differing wavelength arrays. Data that does not overlap with wav will be set to nan. If None,
             this will be set to the wavelengths of the first encounted dataset.
    Returns:
        a HyLibrary spectral library instance.
    """

    files = glob.glob(os.path.join(path, "*/*.txt"))
    libs = {}
    for f in files:
        # spectra name
        n = os.path.splitext(os.path.basename(f))[0]

        # mineral name
        m = os.path.dirname(f).split('/')[-1].split('\\')[-1]

        # load data
        lib = loadLibraryTXT(f)
        if wav is None:
            wav = lib.get_wavelengths()

        # check it is valid and resample to desired range
        delta = [np.min(np.abs(lib.get_wavelengths() - w)) for w in wav]
        if np.min(delta) < hylite.band_select_threshold:
            lib = lib.resample(wav, vb=False, partial=True)
            if m in libs:
                libs[m] = libs[m] + lib  # append
            else:
                libs[m] = lib

    # aggregate
    data = np.full((len(libs), np.max([l.sample_count() for l in libs.values()]), len(wav)), np.nan)
    lib = hylite.HyLibrary(data, lab=list(libs.keys()), wav=wav)
    for i, (k, v) in enumerate(libs.items()):
        lib.data[i, :v.data.shape[0], :] = v.data[:, 0, :]

    return lib

def saveLibraryTXT(path, library):
    """
    Save this library in envii TXT format (see loadLibraryTXT).
    """

    makeDirs(path)

    with open(path, 'w') as f:
        f.write("ENVI ASCII Plot File %s\n" % strftime("[%a %b %d %H:%M:%S %Y]", gmtime()))
        f.write('Column 1: Wavelength\n')

        # gather data and write column headers
        data = [library.get_wavelengths()]
        for i, n in enumerate(library.get_sample_names()):
            f.write('Column %d: %s %s\n' % (i + 2, n, n))
            data.append(np.nanmedian(library[n].data, axis=(0, 1)))  # we need to flatten samples to a single spectra

        # write data block
        for row in np.array(data).T:
            for v in row:
                f.write('  %.6f' % v)
            f.write('\n')
    f.close()

def saveLibraryLIB(path, library):
    path = os.path.splitext(path)[0] + ".lib" # ensure correct file format
    from hylite import io # N.B. this import must be here to avoid circular references
    io.save(path, library.as_image()) # default format is just as an image

def loadLibraryLIB(path):
    from hylite import io  # N.B. this import must be here to avoid circular references
    return io.load(path) # this is handled in the io.load function directly


import unittest
import os
import hylite
from hylite import io
from pathlib import Path
from tempfile import mkdtemp
import shutil
import numpy as np

from hylite.project import Camera, Pushbroom

class TestIO(unittest.TestCase):
    def test_load(self):

        if io.usegdal:
            test = [False, True] # test both GDAL and SPy
        else:
            test = [False] # only test SPy - no gdal
        for gdal in test:
            io.usegdal = gdal
            self.img = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
            self.lib = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"library.csv"))
            self.cld = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"hypercloud.hdr"))

            # test load with SPy
            img = io.loadWithSPy(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
            self.assertTrue( np.nanmax( np.abs(self.img.data - img.data ) ) < 0.01 ) # more or less equal datsets

    def test_loadtxt(self):
        lib = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "library.csv"))
        pth = mkdtemp()
        try:
            io.saveLibraryTXT(os.path.join(pth,"libtxt.txt"), lib )
            io.saveLibraryCSV(os.path.join(pth, "libcsv.csv"), lib)

            lib2 = io.loadLibraryTXT(os.path.join(pth,"libtxt.txt"))
            lib3 = io.loadLibraryCSV(os.path.join(pth, "libcsv.csv"))
            for l in [lib2, lib3]:
                self.assertLess( np.max( np.abs( l.data - lib.data ) ), 1e-5 )
                self.assertLess( np.max(np.abs(l.get_wavelengths() - lib.get_wavelengths())), 1e-5 )

            # test loading from directory
            for i,mineral in enumerate(['quartz', 'biotite','phlogopite']): # build directory
                io.saveLibraryTXT(os.path.join(pth,"library/%s/_%d.txt"%(mineral,i)), lib )
            lib = io.loadLibraryDIR(os.path.join(pth,"library"))
            self.assertIn('phlogopite', lib.get_sample_names())
            self.assertEquals(lib.data.shape[0],3)
            self.assertEquals(lib.data.shape[1],57)
        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - could not load or save spectral library to text format.")

    def test_save(self):
        self.test_load() # load datasets
        pth = mkdtemp()

        if io.usegdal:
            test = [False, True] # test both GDAL and SPy
        else:
            print("Warning - GDAL is not installed. GDAL related functions will not be tested.")
            test = [False] # only test SPy - no gdal

        for gdal in test:
            io.usegdal = gdal
            # test lib and cloud
            try:
                for data,name in zip([self.lib, self.cld],['lib','cld']):
                    io.save(os.path.join(pth, "%s.hdr" % name), data)
                    data2 = io.load(os.path.join(pth, "%s.hdr" % name))
                    self.assertAlmostEquals( np.nanmax(np.abs( data.data - data2.data)), 0, 6 ) # check values are the same

                # test image(s) with GDAL and SPy
                for data in [self.img]:
                        # save with default (GDAL?)
                        io.save(os.path.join(pth, "data.hdr"), data )
                        self.assertEqual( os.path.exists(os.path.join(pth, "data.hdr")), True)
                        data2 = io.load(os.path.join(pth, "data.hdr")) # reload it
                        self.assertAlmostEquals(np.nanmax(np.abs(data.data - data2.data)), 0,
                                                6)  # check values are the same

                        # save with SPy
                        io.saveWithSPy(os.path.join(pth, "data2.hdr"), data )
                        self.assertEqual(os.path.exists(os.path.join(pth, "data2.hdr")), True)
                        data2 = io.load(os.path.join(pth, "data2.hdr")) # reload it
                        self.assertAlmostEquals(np.nanmax(np.abs(data.data - data2.data)), 0,
                                                6)  # check values are the same

                # test saving 3-band images to png files
                rgb = self.img.export_bands(hylite.RGB) # get 3-band image
                rgb.percent_clip(1,99,per_band=True,clip=True) # scale to range 0 - 1
                rgb.data = (rgb.data * 255).astype(np.uint8) # convert to uint8
                io.save(os.path.join(pth, "rgb.hdr"), rgb )
                print(os.listdir(pth))
                self.assertTrue(os.path.exists(os.path.join(pth,'rgb.png')))

                # test loading png image from header file
                img = io.load(os.path.join(pth,'rgb.hdr'))
                self.assertTrue( img is not None)

                # test saving camera objects
                cam = Camera( np.ones(3), np.ones(3), 'pano', 32.2, (100,100), step=0.1 ) # build test camera
                track = Pushbroom( np.ones((1000,3)), np.ones((1000,3)), 0.05, 30.04, (100,1000)) # test pushbroom camera

                # test saving camera
                data = cam
                io.save(os.path.join(pth, "camera"), data )
                self.assertTrue(os.path.exists(os.path.join(pth, "camera.cam")))

                # test saving track
                data = track
                io.save(os.path.join(pth, "track"), data )
                self.assertTrue(os.path.exists(os.path.join(pth, "track.brm")))

                # test loading camera
                cam2 = io.load( os.path.join(pth, "camera.cam"))
                self.assertTrue( (np.abs(cam2.pos - cam.pos) < 0.01).all() )
                self.assertTrue((np.abs(cam2.ori - cam.ori) < 0.01).all())
                self.assertTrue(cam2.dims[0] == cam.dims[0])
                self.assertEqual(cam2.proj, cam.proj)

                # test loading pushbroom
                track2 = io.load(os.path.join(pth, "track.brm"))
                self.assertTrue((np.abs(track2.cp - track.cp) < 0.001).all())
                self.assertTrue((np.abs(track2.co - track.co) < 0.001).all())
                self.assertTrue(track2.dims[0] == track.dims[0])
                self.assertTrue( np.abs(track2.pl - track.pl) < 0.001 )

                # test export to envi
                from hylite.io.libraries import saveLibraryTXT, loadLibraryTXT
                saveLibraryTXT( os.path.join(pth,"lib.txt"), self.lib )
                lib2 = loadLibraryTXT(  os.path.join(pth,"lib.txt") )
                self.assertTrue( (np.abs( self.lib.get_wavelengths() - lib2.get_wavelengths()) < 1e-5).all() )
                self.assertTrue((np.abs(self.lib.data - lib2.data) < 1e-5).all())

                # test save legend
                from hylite.analyse import saveLegend
                saveLegend('Red stuff', 'Green stuff', 'Blue stuff', os.path.join(pth, 'legend.png'))
                self.assertTrue(os.path.exists( os.path.join(pth, 'legend.png') ))

            except:
                shutil.rmtree(pth)  # delete temp directory
                self.assertFalse(True, "Error - could not save data of type %s" % str(type(data)))

            shutil.rmtree(pth)  # delete temp directory

    def test_hycollection(self):

        # load some data
        self.test_load()  # load datasets
        pth = mkdtemp()
        try:
            # build a HyCollection
            C = hylite.HyCollection( name = "testC", root = pth )
            C.img = self.img
            C.cld = self.cld
            C.lib = self.lib
            C.val = 100.
            C.arr = np.linspace(0,100)
            C.x = None # this should be ignored
            C.bool = True

            # save it
            io.save( os.path.join(pth, "testC.hdr"), C )
            self.assertTrue(os.path.exists(os.path.join(os.path.join(pth, "testC.hyc"),"arr.npy")))  # check numpy array has been saved
            self.assertTrue(os.path.exists(os.path.join(os.path.join(pth, "testC.hyc"),"img.hdr")))  # check image has been saved

            # load it
            C2 = io.load( os.path.join(pth, "testC.hdr") )

            # test it
            self.assertEqual( C2.val, C.val )
            self.assertTrue( (C2.arr == C.arr).all() )
            self.assertTrue( C2.bool )
            self.assertEqual(C2.val, C.val)
            self.assertEqual(C2.val, C.val)
            self.assertEqual(C2.img.xdim(), C.img.xdim())
            self.assertEqual(C2.cld.point_count(), C.cld.point_count())
            self.assertEqual(C2.lib.sample_count(), C.lib.sample_count())

            # test cleaning
            C2.bool = None
            C2.val = None
            C2.img = None
            C2.arr = None
            C2.clean()

            self.assertFalse( 'bool' in C2.header )
            self.assertFalse('val' in C2.header)
            self.assertFalse(os.path.exists(os.path.join(os.path.join(pth, "testC.hyc"),"val.npy")))  # check numpy array has been deleted
            self.assertFalse(os.path.exists(os.path.join(os.path.join(pth, "testC.hyc"),"img.hdr"))) # check image has been deleted

            # add a relative path!
            C2.addExternal( 'relobject', os.path.join(os.path.join(pth, "testC.hyc"),"lib.lib") )
            self.assertTrue( isinstance(C2.relobject, hylite.HyLibrary) )

            # test saving collection in a different location
            io.save(os.path.join(pth, "testD.hdr"), C2 )
            self.assertTrue(os.path.exists(os.path.join(os.path.join(pth, "testD.hyc"),"cld.hdr")))  # check cloud has been copied across

            # load this collection
            C3 = io.load(os.path.join(pth, "testD.hyc"))
            C3.inner = io.load( os.path.join(pth, "testC.hdr") ) # add nested collection
            C3.inner.arr2 = np.full( 40, 3.0 ) # add new thing to nested collection
            io.save(os.path.join(pth, "testE.hyc"), C3 )
            self.assertTrue(os.path.exists(os.path.join(os.path.join(pth, "testE.hyc"),"cld.hdr"))) # check cloud has been copied across
            self.assertTrue(os.path.exists(os.path.join(os.path.join(os.path.join(pth,
                                                        "testE.hyc"),"inner.hyc"),"cld.hdr"))) # check cloud has been copied across
            self.assertTrue(os.path.exists(os.path.join(os.path.join(os.path.join(pth,
                                                        "testE.hyc"),"inner.hyc"),"arr2.npy"))) # check cloud has been copied across
            self.assertTrue(isinstance(C3.relobject, hylite.HyLibrary)) # check relative path link can be loaded

            # test quicksave function
            C3.save()
        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - could not create, load or save HyCollection." )
        shutil.rmtree(pth)  # delete temp directory

    def test_subset(self):
        from hylite.io.images import loadSubset

        # load whole image for reference
        path = os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "image.hdr")
        image = io.load(path)

        # load subset and check that dimensions and values match
        subset = loadSubset(path, bands=hylite.SWIR )
        self.assertEqual(subset.xdim(), image.xdim())
        self.assertEqual(subset.ydim(), image.ydim())
        self.assertAlmostEqual(np.nanmax( np.abs(image.export_bands(hylite.SWIR).data - subset.data ) ), 0 )

        # load a pixel and check that the dimensions and values match

if __name__ == '__main__':
    unittest.main()


"""
Read common point cloud formats (.ply, .laz, .csv).
"""

import os
import numpy as np
from hylite import HyCloud
from .headers import matchHeader, makeDirs, saveHeader, loadHeader
from .images import loadWithGDAL


def saveCloudCSV(path, cloud, delimeter=' ', fmt='%.3f', writeHeader=True):

    """
    Saves a HyperCloud object to csv format.

    Args:
        path: the path to save with. If an extension is provided (e.g. .las) then it will be retained. Otherwise .csv is used.
        cloud: the hypercloud to write
        delimeter: the delimeter to use to separate values. Default is ",".
        fmt: numeric value formatting (see documentation for np.savetxt(...) for more details).
        write: True if first row should contain field names (including scalar fields). Default is true.
    """

    # make directories if need be
    makeDirs( path )

    # add file extention to path if need be
    pth, ext = os.path.splitext(path)
    if ext is '': path += ".csv"

    # build header and data arrays
    d = delimeter
    format = [fmt, fmt, fmt]
    data = [cloud.xyz]
    header = 'x%sy%sz' % (d, d)

    #add rgb (as integers)
    if cloud.has_rgb():
        header += '%sr%sg%sb' % (d, d, d)
        format += ['%d', '%d', '%d']
        if cloud.rgb.dtype == np.uint8:
            data.append(cloud.rgb)
        elif cloud.rgb.dtype == np.float64 or cloud.rgb.dtype == np.float32:
            data.append((cloud.rgb * 255).astype(int))
        else:
            print("Warning - unknown data type for RGB colours (%s)" % cloud.rgb.dtype)

    #add normals
    if cloud.has_normals():
        header += '%sNx%sNy%sNz' % (d, d, d)
        format += [fmt, fmt, fmt]
        data.append(cloud.normals)

    #add scalar fields
    if cloud.has_bands():

        #calculate band identifiers
        if cloud.has_wavelengths():
            names = cloud.get_wavelengths()
        elif cloud.has_band_names():
            names = cloud.get_band_names()
        else:
            names = [str(x) for x in range(cloud.band_count())]

        for n in names:
            header += '%s%s' % (d, n)
            format += [fmt]
        data.append(cloud.data)

    #strip header?
    if not writeHeader:
        header = ''
    # write csv with numpy
    np.savetxt(path, np.hstack(data), fmt=format, delimiter=delimeter, header=header, newline='\n')

    #write header file
    if not cloud.header is None:
        hdrpth, _ = os.path.splitext( path )
        cloud.push_to_header()
        saveHeader( hdrpth + ".hdr", cloud.header )

def loadCloudCSV(path, delimiter=' ', order='xyzrgbklm'):

    """
    Loads a point cloud from a csv (or other formatted text) file.

    Args:
        path: the file path
        delimeter: the delimeter used (e.g. ' ', ',', ';'). Default is ' '.
        order: A string defining the order of data in the csv. Each character maps to the following:

        - 'x' = point x coordinate
        - 'y' = point y coordinate
        - 'z' = point z coordinate
        - 'r' = point r coordinate
        - 'g' = point g coordinate
        - 'b' = point b coordinate
        - 'k' = point normal (x)
        - 'l' = point normal (y)
        - 'm' = point normal (z).

        Default is 'xyzrgbklm'.

    Returns:
        A HyCloud instance.
    """

    # look for/load header file if one exists
    header, data = matchHeader(path)
    if not header is None:
        header = loadHeader(header) # load header file

    # load points
    points = np.genfromtxt(data, delimiter=delimiter, skip_header=1)
    assert points.shape[1] >= 3, "Error - dataset must at-least contain x,y,z point coordinates."

    # parse order string
    order = order.lower()
    xyz = [order.find('x'), order.find('y'), order.find('z')]
    assert not -1 in xyz, "Error - order must include x,y,z data."
    rgb = [order.find('r'), order.find('g'), order.find('b')]
    norm = [order.find('k'), order.find('l'), order.find('m')]
    notS = []  # columns that have already been parsed (so are not scalar fields)

    # get xyz
    notS += xyz
    xyz = points[:, xyz]

    # try get rgb and normals
    if max(rgb) < points.shape[1] and not -1 in rgb:
        notS += rgb
        rgb = points[:, rgb]

        # if RGB is 0-255 store as uint8 to minimise memory usage
        if (rgb > 1.0).any():
            rgb = rgb.astype(np.uint8)
    else:
        rgb = None
    if max(norm) < points.shape[1] and not -1 in norm:
        notS += norm
        norm = points[:, norm].astype(np.float32) #store norm as float32 rather than float64 (default)
    else:
        norm = None

        # get remaining data as scalar field
    scal = [i for i in range(points.shape[1]) if i not in notS]
    if len(scal) > 0:
        scal = points[:, scal]
    else:
        scal = None

    # return point cloud
    return HyCloud(xyz, rgb=rgb, normals=norm, bands=scal, header=header)

def saveCloudLAS(path, cloud):
    """
    Write a point cloud object to .las. Note that LAS file formats cannot handle scalar field data.

    Args:
        path: the .las file to save to.
        cloud: a HyCloud instance containing data to save
    """

    try:
        import laspy
    except:
        assert False, "Please install laspy (pip install laspy) to export to LAS."

    # make directories if need be
    makeDirs( path )

    header = laspy.header.Header(point_format=2)
    outfile = laspy.file.File(path, mode='w', header=header)

    # setup header
    outfile.header.offset = [np.min(cloud.xyz[:, 0]),
                             np.min(cloud.xyz[:, 1]),
                             np.min(cloud.xyz[:, 2])]
    outfile.header.scale = [1.0, 1.0, 1.0]
    outfile.header.max = [np.max(cloud.xyz[:,0]), np.max(cloud.xyz[:,1]), np.max(cloud.xyz[:,2])]
    # export point coordinates
    outfile.X = cloud.xyz[:, 0]
    outfile.Y = cloud.xyz[:, 1]
    outfile.Z = cloud.xyz[:, 2]

    # export point colours
    if cloud.has_rgb():
        if cloud.rgb.dtype == np.uint8: #RGB is 0-255
            outfile.red = cloud.rgb[:, 0]
            outfile.green = cloud.rgb[:, 1]
            outfile.blue = cloud.rgb[:, 2]
        elif cloud.rgb.dtype is np.float64 or cloud.rgb.dtype is np.float32: #RGB is 0-1
            outfile.red = (cloud.rgb[:, 0] * 255).astype(int)
            outfile.green = (cloud.rgb[:, 1] * 255).astype(int)
            outfile.blue = (cloud.rgb[:, 2] * 255).astype(int)
        else:
            print("Warning - unknown data type for RGB colours (%s)" % cloud.rgb.dtype)

    outfile.close()

    #write header file
    if not cloud.header is None:
        hdrpth, _ = os.path.splitext( path )
        cloud.push_to_header()
        saveHeader(hdrpth + ".hdr",cloud.header)

def loadCloudLAS(path):
    """
    Loads a LAS file from the specified path.
    """

    try:
        import laspy
    except:
        assert False, "Please install laspy (pip install laspy) to load LAS."

    # look for/load header file if one exists
    header, data = matchHeader(path)
    if not header is None:
        header = loadHeader(header) # load header file

    f = laspy.file.File(data, mode='r')

    # get data
    xyz = np.array([f.x, f.y, f.z], dtype=f.x.dtype).T

    # try to get colour info
    try:
        rgb = np.array([f.get_red(), f.get_green(), f.get_blue()], dtype=f.get_blue().dtype)
    except:
        rgb = None

    # construct cloud
    return HyCloud(xyz, rgb=rgb, header=header)

def saveCloudPLY(path, cloud, sfmt=None):
    """
    Write a point cloud and associated RGB and scalar fields to .ply.

    Args:
        path: the .ply file to save to.
        cloud: a HyCloud instance containing data to save
        sfmt: the format for scalar field data. Can be 'u1', 'u2' or 'f4':

            - 'u1' uses one byte per point per scalar field (255 possible values). This results in smaller file size.
            - 'u2' uses two bytes per point per scalar field (65535 possible values).
            - 'f4' uses four bytes per point per scalar field (float32 precision). This results in large files.

            Default (None) is to use u2 for integer data and f4 for float data.
    """

    # make directories if need be
    makeDirs( path )

    try:
        from plyfile import PlyData, PlyElement
    except:
        assert False, "Please install plyfile (pip install plyfile) to export to PLY."

    # calculate format?
    if cloud.has_bands():
        if sfmt is None:
            if cloud.is_int():
                sfmt='u2'
            else:
                sfmt='f4'

        #estimate size of file
        ps = 4
        if 'u2' in sfmt.lower(): ps = 2
        elif 'u4' in sfmt.lower(): ps = 4
        ps = (3 * 4 + 3 + cloud.band_count() * ps)
        file_size = cloud.point_count() * ps
        if file_size > 10e9:
            print("Warning: writing large point cloud file (%.1f Gb)." % (file_size / 1e9))

    # create structured data arrays
    vertex = np.array(list(zip(cloud.xyz[:, 0], cloud.xyz[:, 1], cloud.xyz[:, 2])),
                      dtype=[('x', 'double'), ('y', 'double'), ('z', 'double')])

    # create ply elements to write to file
    ply = [PlyElement.describe(vertex, 'vertices')]

    if cloud.has_rgb():

        # map to correct data type
        if cloud.rgb.dtype == np.uint8:
            irgb = cloud.rgb
        elif np.nanmax(cloud.rgb) <= 255 and np.nanmin(cloud.rgb) >= 0:
            irgb = cloud.rgb.astype(np.uint8)
        else:
            assert np.nanmin(cloud.rgb) >= 0 and np.nanmax(
                cloud.rgb) <= 1, "Error: rgb must be int (0-255) or float (0-1)."

            irgb = (cloud.rgb * 255).astype(np.uint8)

        # convert to structured arrays
        irgb = np.array(list(zip(irgb[:, 0], irgb[:, 1], irgb[:, 2])),
                        dtype=[('r', 'u1'), ('g', 'u1'), ('b', 'u1')])
        ply.append(PlyElement.describe(irgb, 'color'))  # create ply elements

    if cloud.has_normals():
        # convert to structured arrays
        norm = np.array(list(zip(cloud.normals[:, 0], cloud.normals[:, 1], cloud.normals[:, 2])),
                        dtype=[('x', 'f4'), ('y', 'f4'), ('z', 'f4')])
        ply.append(PlyElement.describe(norm, 'normals'))  # create ply elements

    if cloud.has_bands():

        # map scalar fields to required type
        if 'u1' in sfmt.lower():
            if cloud.data.dtype == np.uint8:
                data = cloud.data  # no need to do any mapping :)
            else:
                # map scalar field to range 0 - 1
                data = cloud.data.astype(np.float32)
                data -= np.nanmin(data)
                data = data / np.nanmax(data)

                # and compress to uint8
                data = (data * 255).astype(np.uint8)

        elif 'u2' in sfmt.lower():
            if cloud.data.dtype == np.uint16:
                data = cloud.data  # no need to do any mapping :)
            else:
                cloud.compress()
                data = cloud.data
        elif 'f4' in sfmt.lower():
            if cloud.data.dtype == np.float32:
                data = cloud.data  # no need to do any mapping :)
            elif cloud.data.dtype == np.float64:
                data = cloud.data.astype(np.float32)
            else:
                cloud.decompress()
                data = cloud.data.astype(np.float32)
        else:
            assert False, "Error - %s is an invalid data format." % sfmt
        # build data arrays
        for b in range(cloud.band_count()):

            #generate band names
            n = str(b)
            if cloud.has_wavelengths():
                n = str(cloud.get_wavelengths()[b])
            elif cloud.has_band_names():
                n = str(cloud.get_band_names()[b])

            #remove spaces from n
            n = n.strip()
            n = n.replace(' ', '_')

            #name already includes 'scalar'?
            if 'scalar' in n:
                ply.append(PlyElement.describe(data[:, b].astype([('%s' % n, sfmt)]), '%s' % n))

            #otherwise prepend it (so CloudCompare recognises this as a scalar field).
            else:
                ply.append(PlyElement.describe(data[:, b].astype([('scalar_%s' % n, sfmt)]), 'scalar_%s' % n))

    # write
    PlyData(ply).write(path)

    #write header file
    if not cloud.header is None:
        hdrpth, _ = os.path.splitext( path )
        cloud.push_to_header()
        saveHeader(hdrpth + ".hdr",cloud.header)

def loadCloudPLY(path):
    """
    Loads a PLY file from the specified path.
    """

    try:
        from plyfile import PlyData, PlyElement
    except:
        assert False, "Please install plyfile (pip install plyfile) to load PLY."

    # look for/load header file if one exists
    header, data = matchHeader(path)
    if not header is None:
        header = loadHeader(header) # load header file

    # load file
    data = PlyData.read(data)

    # extract data
    xyz = None
    rgb = None
    norm = None
    scalar = []
    scalar_names = []
    for e in data.elements:
        if 'vert' in e.name.lower():  # vertex data
            xyz = np.array([e['x'], e['y'], e['z']]).T

            if len(e.properties) > 3:  # vertices have more than just position
                names = e.data.dtype.names

                # colour?
                if 'red' in names and 'green' in names and 'blue' in names:
                    rgb = np.array([e['red'], e['green'], e['blue']], dtype=e['red'].dtype).T

                # normals?
                if 'nx' in names and 'ny' in names and 'nz' in names:
                    norm = np.array([e['nx'], e['ny'], e['nz']], dtype=e['nx'].dtype).T

                # load others as scalar
                mask = ['red', 'green', 'blue', 'nx', 'ny', 'nz', 'x', 'y', 'z']
                for n in names:
                    if not n in mask:
                        scalar_names.append(n)
                        scalar.append(e[n])

        elif 'color' in e.name.lower():  # rgb data
            rgb = np.array([e['r'], e['g'], e['b']], dtype=e['r'].dtype).T
        elif 'normal' in e.name.lower():  # normal data
            norm = np.array([e['x'], e['y'], e['z']], dtype=e['z'].dtype).T
        else:  # scalar data
            scalar_names.append(e.properties[0].name.strip())
            scalar.append(np.array(e[e.properties[0].name], dtype=e[e.properties[0].name].dtype))

    if len(scalar) == 0:
        scalar = None
        scalar_names = None
    else:
        scalar = np.vstack(scalar).T

    assert (not xyz is None) and (xyz.shape[0] > 0), "Error - cloud has not points?"

    # return HyCloud
    return HyCloud(xyz, rgb=rgb, normals=norm, bands=scalar, band_names=scalar_names, header=header)

def loadCloudDEM(pathDEM, pathRGB=None):
    """
    Combines a DEM file and an Orthophoto from the specified path to a cloud.
    """

    # load files
    DEMdata = loadWithGDAL(pathDEM)
    if not pathRGB is None:
        RGBdata = loadWithGDAL(pathRGB)

    # create list of pixel coordinates
    pixgrid = np.meshgrid(np.arange(0,DEMdata.xdim(),1),np.arange(0,DEMdata.ydim(),1))
    pixlist = np.vstack(((pixgrid[0]).flatten(),(pixgrid[1]).flatten()))

    # transform pixel coordinates to world xy coordinates
    fx=[]
    fy=[]
    for i in range(pixlist.shape[1]):
        tx,ty= DEMdata.pix_to_world(int(pixlist[0,i]),int(pixlist[1,i]))
        fx.append(tx)
        fy.append(ty)

    # add DEM data as z information and create xyz cloud list
    fz = DEMdata.data.flatten()
    xyz = np.array([fx, fy, fz]).T

    # if orthophoto is specified, extract RGB values for each DEM point and colorize cloud
    if not pathRGB is None:
        rgb=[]
        for i in range(len(xyz)):
            wx,wy = RGBdata.world_to_pix(xyz[i,0],xyz[i,1])
            rgb.append(RGBdata.data[wx,wy,:])

    # otherwise fill rgb info with zeros
    else:
        rgb = np.zeros((len(xyz),3))

    # return HyCloud
    return HyCloud(xyz, rgb=rgb)




import unittest
from tempfile import mkdtemp
import shutil
import numpy as np
import hylite
from hylite import io
import os
from pathlib import Path

class MyTestCase(unittest.TestCase):
    def getTestData(self):
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "image.hdr"))
        image.data[:50, :, :] = np.nan  # add some nans to make more realistic
        cloud = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "hypercloud.hdr"))

        # also create a HyLibrary instance
        image.header.set_sample_points('A', [(20, 15)])  # label some seed pixels in each sample
        image.header.set_sample_points('B', [(80, 15)])
        image.header.set_sample_points('C', [(140, 15)])
        image.header['class names'] = ['A', 'B', 'C']
        from hylite.hylibrary import from_indices
        lib = from_indices(image,
                           [image.header.get_sample_points(n)[0] for n in image.header.get_class_names()],
                           names=image.header.get_class_names(),
                           s=5)
        return image, cloud, lib

    def test_save_header_only(self):
        from hylite import HyCollection
        pth = mkdtemp()  # create output directory
        try:
            C = HyCollection("test", pth ) # create a HyCollection
            C.attr = "foo"
            self.assertEquals(C.file_type, 'Hylite Collection') # check "file type" key is loaded as file_type
            self.assertEquals(C.file_type, C.header['file type'])  # check "file type" key is loaded as file_type
            C.save()

            C2 = io.load(os.path.join(pth,'test.hdr'))
            self.assertEquals(C.attr, "foo")
            self.assertEquals(C2.file_type, 'Hylite Collection')  # check "file type" key is loaded as file_type
            self.assertEquals(C2.file_type, C2.header['file type'])  # check "file type" key is loaded as file_type
        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - failed save header only test")
        shutil.rmtree(pth)  # delete temp directory



    def test_basic(self):
        from hylite import HyCollection
        pth = mkdtemp() # create output directory
        array = np.random.rand(50)  # create some random data
        image, cloud, lib = self.getTestData()
        try:
            ### Create and save a HyCollection
            C = HyCollection("test", pth ) # create a HyCollection
            C.array = array # put in numpy array
            C.image = image
            C.cloud = cloud
            C.lib = lib

            X = C.cloud # call getter
            # check that the directory does not exist yet!
            print( C.getDirectory(makedirs=False) )
            self.assertFalse( os.path.exists( C.getDirectory(makedirs=False) ) )
            # test get and set
            self.assertEqual( C.get('image'), image )
            C.set('image2', image)
            self.assertEqual(C.image2, image )

            # test loaded function
            self.assertTrue( C.loaded('image2'), "Error in loaded(...) function.")
            # test save
            C.save_attr('image2') # save image attribute
            C.set('image3', image, save=True) # save on set
            C.save() # save all attributes

            ### Reload it
            C2 = io.load( os.path.join(pth,'test.hdr'))
            self.assertFalse( C2.loaded('image2'), "Error in loaded(...) function.")

            # check getAttributes() function
            self.assertTrue( len(C2.getAttributes(True)) == 0, "Error - getAttributes(ram=True) returned attributes on disk." )
            self.assertEqual( len(C2.getAttributes(False)), len(C.getAttributes(False) )), "Error - getAttributes(ram=True) did not return attributes on disk."
            # print(C.getAttributes(False))
            # print(C2.getAttributes(False))

            # check equality
            thresh = 1e-5
            self.assertTrue( np.max( np.abs( C2.array - array ) ) < thresh )
            self.assertTrue(np.nanmax(np.abs(C2.image.data - image.data)) < thresh)
            self.assertTrue(np.nanmax(np.abs(C2.cloud.data - cloud.data)) < thresh)
            self.assertTrue(np.nanmax(np.abs(C2.lib.data - lib.data)) < thresh)
        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - failed basic HyCollection tests")
        shutil.rmtree(pth)  # delete temp directory

    def test_image_png(self):
        from hylite import HyCollection
        pth = mkdtemp()  # create output directory
        array = np.random.rand(50)  # create some random data
        image, cloud, lib = self.getTestData()
        try:
            C = HyCollection("test", pth)  # create a HyCollection
            C.image = image
            C.rgb = image.export_bands(hylite.RGB)
            C.rgb.percent_clip()
            C.rgb.header['magic_key'] = '42'
            C.rgb.data = (C.rgb.data * 255).astype(np.uint8) # convert to uint8 - this should be saved as .png
            C.save()
            C.free()

            self.assertTrue(os.path.exists(os.path.join(C.getDirectory(), "rgb.hdr"))) # check header file written
            self.assertTrue( os.path.exists( os.path.join( C.getDirectory(), "rgb.png" ) ) ) # this should be png
            self.assertTrue(os.path.exists(os.path.join(C.getDirectory(), "image.dat"))) # this should be dat
            self.assertEqual( C.rgb.data.dtype, np.uint8 ) # load from disk and check type
            self.assertListEqual(list(C.rgb.get_wavelengths()), [503.4, 551.19, 681.63] ) # check header info is preserved
            self.assertEqual(C.rgb.header['magic_key'], '42')

        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - failed nested HyCollection tests")
        shutil.rmtree(pth)  # delete temp directory

    def test_nested(self):
        from hylite import HyCollection
        pth = mkdtemp() # create output directory
        array = np.random.rand(50)  # create some random data
        image, cloud, lib = self.getTestData()
        try:
            ### Create and save a HyCollection
            C = HyCollection("test", pth ) # create a HyCollection
            C.funky_data_base = 'foobar'
            C.addSub( "SC1" ) # create a subcollection
            C.SC1.array = array
            C.SC1.image = image
            C.SC1.funky_data_A = array

            SC2 = C.addSub( "SC2" ) # add another subcollection
            SC2.cloud = cloud
            SC2.lib = lib
            SC2.funky_data_B = lib

            C.addSub("SC3") # add a subcollection with no data directory (only header)
            C.SC3.message = "hi!"
            C.SC3.funky_data_C = 'This is the answer!'
            C.SC3.another_image = image.export_bands(hylite.RGB)
            C.save()

            ### Reload it
            C2 = io.load( os.path.join(pth,'test.hdr'))

            # check equality
            thresh = 1e-5
            self.assertTrue( np.max( np.abs( C2.SC1.array - array ) ) < thresh )
            self.assertTrue(np.nanmax(np.abs(C2.SC1.image.data - image.data)) < thresh)
            self.assertTrue(np.nanmax(np.abs(C2.SC2.cloud.data - cloud.data)) < thresh)
            self.assertTrue(np.nanmax(np.abs(C2.SC2.lib.data - lib.data)) < thresh)
            self.assertTrue('hi' in C2.SC3.message )

            # test recursive searching
            C2.free() # used for testing queries on datasets that are not loaded into RAM
            for _C in [C, C2]:
                query = _C.query(name_pattern='funky_data', recurse = True, ram_only=False)
                self.assertListEqual(sorted(query), sorted(['funky_data_base', 'funky_data_B', 'funky_data_A', 'funky_data_C']) )
                query = _C.query(name_pattern='funky_data', recurse=False, ram_only=False)
                self.assertListEqual(query, ['funky_data_base']) # only matches a single argument
                query = _C.query(ext_pattern=['npy', 'ndarray'], recurse=True, ram_only=False) # matches string attributes
                self.assertListEqual(query, sorted(['array', 'funky_data_A']))
                query = _C.query(ext_pattern=['dat', 'HyImage'], recurse=True,
                                 ram_only=False)  # matches string attributes
                self.assertListEqual(query, ['another_image', 'image'])
                query = _C.query(ext_pattern=['hyc', 'HyCollection'], recurse=False, ram_only=False)  # test no recurse
                query2 = _C.query(ext_pattern=['hyc', 'HyCollection'], recurse=True, ram_only=False,
                                  recurse_matches=False )  # test no recurse matches
                self.assertListEqual(query, query2)
        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - failed nested HyCollection tests")
        shutil.rmtree(pth)  # delete temp directory

if __name__ == '__main__':
    unittest.main()



import unittest
import os
from pathlib import Path
import hylite.io as io
from hylite.project.align import *
import shutil
from tempfile import mkdtemp
import numpy as np

class TestAlign(unittest.TestCase):
    def test_warp(self):
        # load test image
        image1 = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        image1.data = np.dstack( [ image1.data, image1.data ] ) # make dataset > 512 bands

        # create slightly offset second image
        image2 = image1.copy()
        image2.data = image2.data[5:, 5:, : ].copy() # add an offset
        image2.push_to_header()

        # test align with method=affine
        # n.b. there's not really any way to check if the align actually worked... so just try a bunch of combinations
        # np.warnings.filterwarnings('ignore') # supress warnings when comparing to nan
        align_images( image1, image2, image_bands=(30, 50), target_bands=(30, 50),
                                warp=True, features='sift', method='affine' )
        align_images(image1, image2, image_bands=(450., 600.), target_bands=(450., 600.),
                     warp=True, features='sift', method='poly')

        # test piecewise align
        for m in ['piecewise', 'affine', 'polynomial']:
            align( image1, image2, source_bands=(0,1,2), method=m )

    def test_align_to_cloud(self):

        # generate a basic geometry
        x, y = np.meshgrid(np.linspace(-10, 10), np.linspace(-10, 10))
        xyz = np.vstack([x.ravel(), y.ravel(), np.zeros_like(x.ravel())]).T
        cloud = hylite.HyCloud( xyz )

        # project onto a camera
        img_dims = (20, 20)
        cam = Camera(np.array([0, 0, 40]), np.array([0, 0, 90]), 'persp', 25., dims=img_dims)
        pp, vis = proj_persp(xyz, cam.pos, cam.ori, fov=cam.fov, dims=cam.dims)

        # choose 20 random points
        idx = [np.random.randint(0, np.sum(vis)) for i in range(20)]
        pixels = pp[vis,:][idx, :2 ]
        points = np.arange(xyz.shape[0])[vis][idx]
        cam_est, err = align_to_cloud_manual(cloud, cam, points, pixels)

        #print(err, np.linalg.norm(cam.pos - cam_est.pos), np.linalg.norm(cam.ori - cam_est.ori))
        self.assertAlmostEqual(np.linalg.norm(cam.pos - cam_est.pos), 0, 3 )
        self.assertAlmostEqual( np.sin( np.deg2rad( np.linalg.norm(cam.ori - cam_est.ori) ) ), 0, 3 )


if __name__ == '__main__':
    unittest.main()



"""
Put .txt files for custom reference spectra in here ( each line should contain a wavelength, reflectance pair) and they
will be loaded automatically as global variables. For example *MySpectra.txt* would become accessible as
*hylite.reference.spectra.MySpectra*.
"""


"""
Generic tools for distributing computationally intensive tasks across multiple threads.
"""

import os
import numpy as np
import shutil
from tempfile import mkdtemp
import multiprocessing as mp

from hylite import HyCloud, HyImage
from hylite import io


def _split(data, nchunks):
    """
    Split the specified HyCloud instance into a number of chunks.

    Args:
        data = the complete HyData object to copy and split.
        nchunks = the number of chunks to split into.
    Returns:
        a list of split
    """

    if isinstance(data, HyCloud):  # special case for hyperclouds - split xyz, rgb and normals too
        chunksize = int(np.floor(data.point_count() / nchunks))
        chunks = [(i * chunksize, (i + 1) * chunksize) for i in range(nchunks)]
        chunks[-1] = (chunks[-1][0], data.point_count())  # expand last chunk to include remainder

        # split points
        xyz = [data.xyz[c[0]:c[1], :].copy() for c in chunks]

        # split data
        bands = [None for c in chunks]
        if data.has_bands():
            X = data.get_raveled().copy()
            bands = [X[c[0]:c[1], :] for c in chunks]

        # split rgb
        rgb = [None for c in chunks]
        if data.has_rgb():
            rgb = [data.rgb[c[0]:c[1], :].copy() for c in chunks]

        # split normals
        normals = [None for c in chunks]
        if data.has_normals():
            normals = [data.normals[c[0]:c[1], :].copy() for c in chunks]

        return [HyCloud(xyz[i],
                        rgb=rgb[i],
                        normals=normals[i],
                        bands=bands[i],
                        header=data.header.copy()) for i in range(len(chunks))]

    else:  # just split data (for HyImage and other types)
        X = data.get_raveled().copy()
        chunksize = int(np.floor(X.shape[0] / nchunks))
        chunks = [(i * chunksize, (i + 1) * chunksize) for i in range(nchunks)]
        chunks[-1] = (chunks[-1][0], X.shape[0])  # expand last chunk to include remainder

        out = []
        for c in chunks:
            _o = data.copy(data=False)  # create copy
            _o.data = X[c[0]:c[1], :][:,None,:]
            out.append(_o)

        return out

def _merge(chunks, shape):
    """
    Merge a list of HyData objects into a combined one (aka. do the opposite of split(...)).

    Args:
        chunks = a list of HyData chunks to merge.
        shape = the output data shape.
    Returns: a single merged HyData instance (of the same type as the input).
               The header of this instance will be a copy of chunks[0].header.
    """

    # merge data
    X = np.vstack([c.data for c in chunks])
    X = X.reshape((*shape, -1))

    if not isinstance(chunks[0], HyCloud): # easy!
        # make copy
        out = chunks[0].copy(data=False)
        out.data = X
        out.header = chunks[0].header.copy()
        return out
    else: # less easy
        xyz = np.vstack([c.xyz for c in chunks])
        rgb = None
        if chunks[0].has_rgb():
            rgb = np.vstack([c.rgb for c in chunks])
        normals = None
        if chunks[0].has_normals():
            normals = np.vstack([c.normals for c in chunks])
        return HyCloud( xyz, rgb=rgb, normals=normals, bands=X, header=chunks[0].header.copy())

def _call(func, path, arg, kwd, n):
    """
    This function will be called by each thread. It loads each data chunk from disk, runs the operation, then saves
    the results.
    """

    # print("Spawning thread %d." % n)
    # func, path, arg, kwd = args

    # load data chunk
    if '.ply' in path:
        data = io.loadCloudPLY(path)  # load point cloud
        result = func(data, *arg, **kwd)  # compute results
        assert isinstance(result, HyCloud), "Error - function %s does not return a HyCloud." % func
        io.saveCloudPLY(path, result)  # save point cloud
    else:
        data = io.load(path)  # load image
        result = func(data, *arg, **kwd)  # compute results
        assert isinstance(result, HyImage), "Error - function %s does not return a HyImage." % func
        io.save(path, result)  # save result

    return True  # done

def parallel_chunks(function, data, *args, **kwds):

    """
    Run a function that operates per-point or per-pixel on smaller chunks of a point cloud or image dataset
    in parallel. Only use for expensive operations as otherwise overheads (writing files to cache, spawning threads,
    loading files from cache) are too costly.

    Args:
        function: the function to run on each chunk of the dataset. Must take a HyCloud or HyImage dataset as it's first
                  argument and also return a HyCloud or HyImage dataset (cf., mwl(...), get_hull_corrected(...)).
        data (HyCloud or HyImage): data to run the function on.
        args (tuple): tuple of arguments to pass to the function.
        nthreads (int): the number of threads to spawn. Default is the number of cores - 2. Negative numbers will be subtracted
                   from the number of cores.
        **kwds: any other keywords are passed to the function
    Returns:
        Nothing
    """
    assert isinstance(data, HyCloud) or isinstance(data, HyImage)

    # get number of threads
    if 'nthreads' in kwds:
        nthreads = kwds['nthreads']
        del kwds['nthreads']
    else:
        nthreads = -2
    if nthreads < 1:
        nthreads = os.cpu_count() - nthreads
    assert nthreads > 0, "Error - cannot spawn %d threads" % nthreads

    assert isinstance(nthreads, int), "Error - nthreads must be an integer."
    assert nthreads is not None, "Error - could not identify CPU count. Please specify nthreads keyword."

    # split data into chunks
    shape = data.data.shape[:-1]  # store shape (important for images)
    chunks = _split(data, nthreads)

    # dump chunks into temp directory
    pth = mkdtemp()  # make temp directory
    print("Writing thread cache to %s:" % pth)

    # dump clouds to directory
    paths = []
    for i, c in enumerate(chunks):

        if isinstance(c, HyCloud):
            p = os.path.join(pth, '%d.ply' % i)
            io.saveCloudPLY(p, c)
        else:
            p = os.path.join(pth, '%d.hdr' % i)
            io.save(p, c)
        paths.append(p)

    # make sure we don't multithread twice when using advanced scipy/numpy functions...
    os.environ['MKL_NUM_THREADS'] = '1'
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['MKL_DYNAMIC'] = 'FALSE'

    # spawn worker processes
    P = [mp.Process(target=_call, args=(function, p, args, kwds, i)) for i, p in enumerate(paths)]
    try:
        for p in P:
            p.start()
        for p in P:
            p.join()

        # successs! load data again...
        if isinstance(data, HyCloud):
            chunks = [io.loadCloudPLY(p) for p in paths]
        else:
            chunks = [io.load(p) for p in paths]

        # remove temp directory
        shutil.rmtree(pth)  # delete temp directory
        print("Process complete (thread cache cleaned successfully).")
    except (KeyboardInterrupt, SystemExit) as e:
        print("Job cancelled. Cleaning temp directory... ", end='')
        shutil.rmtree(pth) # delete temp directory
        print("Done.")
        assert False, "Multiprocessing job cancelled by KeyboardInterrupt or SystemExit."
    except Exception as e:
        print("Error thrown. Cleaning temp directory... ", end='')
        shutil.rmtree(pth)  # delete temp directory
        print("Done.")
        raise e

    # re-enable scipy/numpy multithreading
    del os.environ['MKL_NUM_THREADS']
    del os.environ['OMP_NUM_THREADS']
    del os.environ['MKL_DYNAMIC']

    # merge back into one dataset
    out = _merge(chunks, shape=shape)
    return out

def _call2(func, in_paths, out_paths, kwd, n):

    for i, o in zip(in_paths, out_paths):  # loop through paths managed by this thread
        func(i, o, **kwd)  # call function

def parallel_datasets(function, in_paths, out_paths=None, nthreads=-2, **kwds):
    """
    Parallelise a single function across many HyData datasets.

    Args:
        function: the function to run on each dataset. This should take an input path (string) as its first input and
                  output path (also string) as its second output. Anything returned by the function will be ignored.
        in_paths (list): a list of input paths, each of which will be passed to function in each thread.
        out_paths(list): a list of corresponding output paths that each function should write to. Defaults to in_paths.
        nthreads (int): the number of threads to spawn. Default is the number of cores - 2. Negative numbers are subtracted
                  from the total number of cores.
        **kwds: any keywords are passed directly to function in each thread.
    Returns:
        Nothing
    """

    assert isinstance(in_paths, list), "Error - in_paths must be a list of file paths (string)."

    if out_paths is None:
        out_paths = in_paths
    assert isinstance(out_paths, list), "Error - out_paths must be a list of file paths (string)."
    assert len(out_paths) == len(in_paths), "Error - length of input and output paths must match."

    # get number of threads
    assert isinstance(nthreads, int), "Error - nthreads must be an integer."
    if nthreads < 1:
        nthreads = os.cpu_count() - nthreads
    assert nthreads > 0, "Error - cannot spawn %d threads" % nthreads

    # distribute input paths across threads
    nthreads = min( len(in_paths), nthreads ) # avoid case where we have more threads than paths
    stride = int( len(in_paths) / nthreads )
    inP = []
    outP = []
    for i in range(nthreads):
        idx0 = i*stride
        idx1 = min( (i+1)*stride, len(in_paths) )
        inP.append( in_paths[idx0:idx1] )
        outP.append( out_paths[idx0:idx1] )
    for i in range(len(in_paths) % nthreads): # and add remainder
        inP[i].append(in_paths[-i-1])
        outP[i].append(out_paths[-i - 1])

    # make sure we don't multithread twice when using advanced scipy/numpy functions...
    os.environ['MKL_NUM_THREADS'] = '1'
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['MKL_DYNAMIC'] = 'FALSE'

    # spawn worker processes and wait for jobs to finish
    P = [mp.Process(target=_call2, args=(function, inP[i], outP[i], kwds, i)) for i in range(nthreads)]
    for p in P:
        p.start()
    for p in P:
        p.join()

    # re-enable scipy/numpy multithreading
    del os.environ['MKL_NUM_THREADS']
    del os.environ['OMP_NUM_THREADS']
    del os.environ['MKL_DYNAMIC']



"""
A base class for all types of hyperspectral data. Inherited by HyCloud, HyImage and HyLibrary.
"""

import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt
from scipy import ndimage, signal
import re
import hylite
from hylite import HyHeader
import hylite.reference.features as ref
from hylite.hyfeature import HyFeature, MultiFeature, MixedFeature
from matplotlib.ticker import AutoMinorLocator

from tqdm import tqdm

class HyData(object):

    """
    A generic class for encapsulating hyperspectral (points and images), and associated metadata such as georeferencing, bands etc.

    This class is based around a data numpy array containing the hyperspectral data in a representation such that the last dimension
    corresponds to individual bands (e.g. data[pointID, band] or data[px,py,band]). Note that the data array can be empty!
    """

    @classmethod
    def to_grey(cls, data):
        """
        Return a copy of the specified data array as uint8 greyscale. Useful for OpenCV operations, compression or mapping to RGB.
        """

        return np.uint8(255 * (data - np.nanmin(data)) / (np.nanmax(data) - np.nanmin(data)))

    #####################################
    ##  Instance methods
    #####################################
    def __init__(self, data, header=None, **kwds):
        """
        Create an image object from a data array.

        Args:
            data (ndarray): array such that the last dimension corresponds to individual bands (e.g. data[pointID, band] or data[px,py,band])
            header (hylite.HyHeader): associated header file. Default is None (create a new header).
        """

        #copy reference to data. Note that this can be None!
        self.data = data
        if not data is None:
            self.dtype = data.dtype
        else:
            self.dtype = None

        # header data
        self.set_header(header)

    def __getitem__(self, key):
        """
        Expose underlying data array when using [ ] operators
        """
        return self.data.__getitem__(key)

    def __setitem__(self, key, value):
        """
        Expose underlying data array when using [ ] operators
        """
        self.data.__setitem__(key, value)

    def copy(self, data=True):
        """
        Make a deep copy of this image instance.
        Args:
            data (bool): True if a copy of the data should be made, otherwise only copy header.

        Returns:
            a new HyData instance.
        """

        if not data or self.data is None:
            return HyData( None, header=self.header.copy())
        else:
            return HyData( self.data.copy(), header=self.header.copy())

    def set_header(self, header):
        """
        Loads associated header data into self.header.

        Args:
            header (hylite.HyHeader): a HyHeader object or None.
        """

        #no header - create one
        if header is None:
            self.header = HyHeader()
            return

        #set the header
        self.header = header

    def push_to_header(self):
        """
        Update header data to match this hyperspectral data.
        """
        self.header['samples'] = str(self.samples())
        self.header['bands'] = str(self.band_count())
        self.header['lines'] = str(self.lines())


    #############################################
    ##Expose important parts of the header file
    #############################################
    def has_wavelengths(self):
        """
        True if the header data contains wavelength information.
        """
        return self.header.has_wavelengths()

    def get_wavelengths(self):
        """
        Get the wavelength that corresponds with each band of this image, as stored in the .hdr file.
        """
        if self.has_wavelengths():
            return self.header.get_wavelengths() # return wavelengths
        else:
            return np.arange(self.band_count()) # return band indices as proxy

    def has_band_names(self):
        """
        True if the header data contains band names.
        """
        return self.header.has_band_names()

    def get_band_names(self):
        """
        Return band names as defined in the header file.
        """
        return self.header.get_band_names()

    def has_fwhm(self):
        """
        True if the header data contains band names.
        """
        return self.header.has_fwhm()

    def get_fwhm(self):
        """
        Return band names as defined in the header file.
        """
        return self.header.get_fwhm()

    def set_wavelengths(self, wavelengths):
        """
        Set the wavelengths associated with this hyperspectral data
        """
        if not wavelengths is None:
            assert len(wavelengths) == self.band_count(), "Error - wavelengths must be specified for each band."
        self.header.set_wavelengths(wavelengths)

    def set_band_names(self, names ):
        """
        Set the band names associated with this hyperspectral data
        """
        if not names is None:
            assert len(names) == self.band_count(), "Error - band names must be specified for each band."
        self.header.set_band_names( names )

    def set_fwhm(self, fwhm):
        """
        Set the band widths associated with this hyperspectral data
        """
        if not fwhm is None:
            assert len(fwhm) == self.band_count(), "Error - wavelengths must be specified for each band."
        self.header.set_fwhm(fwhm)

    def is_image(self):
        """
        Return true if this dataset is an image (i.e. data array has dimension [x,y,b]).
        """
        if self.data is None: # for point clouds data can be none
            return False
        return len(self.data.shape) == 3

    def is_point(self):
        """
        Return true if this dataset is an point cloud or related dataset (i.e. data array has dimension [idx,b]). Note
        that this will return true for spectral libraries and other 'cloud like' datasets.
        """
        if self.data is None: # for point clouds data can be none
            return True
        return len(self.data.shape) == 2

    def is_classification(self):
        return 'classification' in self.header['file type'].lower()

    ###################################
    ## Data dimensions and properties
    ###################################
    def band_count(self):
        """
        Return the number of bands in this dataset.
        """
        if self.data is None: return 0
        return self.data.shape[-1]

    def samples(self):
        """
        Return number of samples in this dataset. For 2D data (images) this is the image height (number of pixels in a line
        scanner). For 1D data (point clouds) this is 1 (each point is like an individual sample).
        """
        if self.data is None:
            return 0
        return self.data.shape[0]

    def lines(self):
        """
        Return number of lines in this dataset. For 2D data (images) this is the image height. For 1D data (point clouds)
        this is 1.
        """

        if self.data is None:
            return 0
        if len(self.data.shape) > 2:
            return self.data.shape[1]
        else:
            return 1

    def is_int(self):
        """
        Return true if this dataset contains data with integer precision.
        """
        return np.issubdtype( self.data.dtype, np.integer )

    def is_float(self):
        """
        Return true if this dataset contains data with floating point precision.
        """
        return np.issubdtype( self.data.dtype, np.floating )

    #############################
    ## masking and band removal
    #############################
    def export_bands(self, bands ):
        """
        Export a specified band range to a new HyData instance.

        Args:
            bands (tuple, list): either:
                     (1) a tuple containing the (min,max) wavelength to extract. If range is a tuple, -1 can be used to specify the
                         last band index.
                     (2) a list of bands or boolean mask such that image.data[:,:,range] is exported to the new image.
        """

        # wrap individual integers or floats in a list
        if isinstance(bands, int) or isinstance(bands, float):
            bands = [bands]  # wrap in list

        # calculate bands to remove
        mask = np.full( self.band_count(), True )
        if isinstance(bands, np.ndarray) and bands.dtype == bool:
            # mask is a numpy array containing True for bands to keep, so flip it to get bands to remove.
            mask = np.logical_not(bands)
        elif isinstance(bands, tuple) and len(bands) == 2:
            # get indices of bands to keep and flag these as False in mask.
            if bands[0] == -1: bands = (0, bands[1])
            if bands[1] == -1: bands = (bands[0], self.band_count())
            mn = self.get_band_index(bands[0])
            mx = self.get_band_index(bands[1])

            # calculate mask
            mask[ mn:(mx+1) ] = False
        else:
            # check bands are all indices and flag these bands as False in mask.
            bands = list(bands)
            for i, b in enumerate(bands):
                if isinstance(b,float) or isinstance(b,str):
                    bands[i] = self.get_band_index(b)

            mask[bands] = False

        # check that we're leaving at least one band....
        assert (mask==False).any(), "Error - cannot export image with no bands."

        # quick exit if we are exporting all bands
        if (mask==False).all(): # no bands are masked
            return self.copy()

        # copy this dataset
        subset = self.copy()
        subset.header.drop_bands(mask) # apply mask to header data
        subset.data = np.zeros( self.data[..., np.logical_not(mask) ].shape, dtype=self.data.dtype )
        subset.data[...] = self.data[..., np.logical_not(mask)]

        # special case for spectral libraries
        if hasattr(self, 'upper'):
            if self.upper is not None:
                subset.upper = self.upper[..., np.logical_not(mask)]
        if hasattr(self, 'lower'):
            if self.lower is not None:
                subset.lower = self.lower[..., np.logical_not(mask)]

        return subset

    def delete_nan_bands(self, inplace=True):
        """
        Remove bands in this image that contain only nans.

        Args:
            inplace (bool): True if this operation should be applied to the data in situ. Default is True.

        Returns:
            an image copy with the nan bands removed IF inplaces is False. Otherwise the image is modified inplace.
        """

        if len(self.data.shape) == 3: #hyImage
            cpy = self.export_bands(np.isfinite(self.data).any(axis=(0, 1)))  # remove bands that are all nan
        else: #hycloud
            assert len(self.data.shape) == 2, "Weird error?"
            cpy = self.export_bands(np.isfinite(self.data).any(axis=0))  # remove bands that are all nan
        if inplace:
            self.header = cpy.header
            self.data = cpy.data
        else:
            return cpy

    def set_as_nan(self, value, strict=True):
        """
        Sets data with the specified value to NaN. Useful for handling no-data values.

        Args:
            value (float, int): the value to (permanently) replace with np.nan.
            strict (bool): True if all bands must have this value to set it as nan. Default is True. If False, all occurences of
                    value will be replaced with nan.
        """

        if strict:
            if self.is_int():
                nan = int(self.header.get("data ignore value", 0))
                self.data[ (self.data == value).all(axis=-1) ] = nan
                self.header["data ignore value"] = str(nan)
            else:
                self.data[ (self.data == value).all(axis=-1) ] = np.nan
        else:
            if self.is_int():
                nan = int(self.header.get("data ignore value", 0))
                self.data[ self.data == value ] = nan
                self.header["data ignore value"] = str(nan)
            else:
                self.data[self.data == value] = np.nan

    def mask_bands(self, mn, mx=None, val=np.nan):
        """
        Masks a specified range of bands, useful for removing water features etc.

        Args:
            min (float): the start of the band mask (as per get_band_index(...)).
            max (float): the end of the band mask (as per get_band_index( ... )). Can be None to mask individual bands. Default is
                 None.
            val (float, int): the value to set masked bands to. Default is np.nan. Set to None to keep values but flag bands in band band list.
        """

        if mx is None:
            self.data[..., self.get_band_index(mn)] = np.nan
            return
        elif mx == -1:
            mx = self.band_count()

        # update bad band list
        bbl = np.full( self.band_count(), True )
        if self.header.has_bbl():
            bbl = self.header.get_bbl()
        bbl[..., self.get_band_index(mn): self.get_band_index(mx)] = False
        self.header.set_bbl( bbl )

        if not val is None:
            self.data[..., self.get_band_index(mn): self.get_band_index(mx)] = val

    def mask_water_features(self, mask=None):
        """
        Removes typical water features. By default this removes bands between:

          - 960 - 990 nm
          - 1320 - 1500 nm
          - 1780 - 2050 nm
          - 2400 - 2500 nm

        Custom wavelengths can be set using the mask keyword.

        Args:
            mask (list,tuple): mask custom bands. This should be a list of tuple band indices or wavelengths containing the
                    minimum and maximum wavelenght/index of each region to mask.
        """

        default = [(960.0, 990.0), (1320.0, 1500.0), (1780.0, 2050.0), (2400.0, 2502.0)]
        if mask is None:
            mask = default

        # mask bands
        for mn, mx in mask:
            try:
                self.mask_bands(mn, mx)
            except:
                pass  # ignore errors associated with out of range etc.

    #########################
    ## band getters/setters
    #########################
    def get_band(self, b):
        """
        Gets an individual band from this dataset. If an integer is passed it is treated as a band index. If a string is passed it is
        treated as a band name. If a float is passed then the closest band to this wavelength is retrieved.

        Args:
            b (int,float,str): the band to get. Integers are treated as indices, strings as band names and floats as wavelengths.

        Returns:
            a sliced np.array exposing the band. Note that this is NOT a copy.
        """

        return self.data[...,self.get_band_index(b)]

    def get_band_grey(self, b):
        """
        Returns the specified band as a uint8 greyscale image compatable with opencv.
        """

        return HyData.to_grey( self.get_band(b) )

    def get_raveled(self):
        """
        Get the data array as a 2D array of points/pixels. NOTE: this is just a view of the original data array, so any
        operations changes made to it will affect the original image. Useful for fast transformations!

        Returns:
            pixels (ndarray): an array such that pixel[n][band] gives the spectra of the nth pixel.
        """

        return self.data.reshape(-1, self.data.shape[-1])

    def X(self, onlyFinite = False):
        """
        A shorthand way of writing get_raveled(), as X is conventionally used for a vector of spectra.

        Args:
            onlyFinite (bool): True if data points containing nan bands should be removed from the feature vector. Default is False.
        """
        X = self.get_raveled()
        if onlyFinite:
            return X[ np.isfinite(X).all(axis=-1) ]
        else:
            return X

    def eval(self, op : str, print=False ):
        """
        Evaluate an arithmetic expression on the data array of this HyData instance.

        Args:
            op: The operation string to evaluate. This should follow the following syntax:
                - all numbers are treated as wavelengths (and are passed as floats to get_band_index(...).
                - band indices should be prefixed by `b`. E.g., `b10`.
                - constants should be prefixed by the `$` character, e.g., `$2`.
                - standard python syntax should be used for basic arithmetic (`+`, `-`, `*`, `/`/, `**`).
                - the `:` operator can be used to average values over a band range, e.g., `b10:b15 or `2190:2210`.
                - the `|` operator can be used to include multiple operations that are evaluated to different output
                  bands (e.g. ` b10+b9 | b12:b15 | b5/b6` will result in a 3-band output.
            print: Optionally print the python expression(s) that is/are generated for debugging or other reference. Default is False.

            N.B. This function works by converting the operation string into python code including numpy operations
            on the data array (`a`). Hence other python or numpy (`np`) syntax may (but is not guaranteed to) also work.

        Returns:
            A copy of this HyData instance but with an updated (single-band) data array.
        """

        if 'import' in op: # don't let people do anything toooo wild!
            assert False, "Don't be naughty now!"

        # strip spaces (but keep one at start)
        op = ' ' + op.strip()
        if '|' in op:
            ops = op.split('|')
            out = [self.eval(o, print=print) for o in ops]
            if self.is_image():
                out[0].data = np.dstack([o.data for o in out])
            else:
                out[0].data = np.hstack([o.data for o in out])
            return out[0]
        else:
            # convert non-constants to band indices
            bidx = re.findall('.[0-9.]+', op)
            for s in bidx:
                if s[0] == 'b':  # specified as band index already
                    continue
                if s[0] == '$':  # constant - skip for now
                    continue
                else:
                    ix = self.get_band_index(float(s[1:]))
                    op = op.replace(s[1:], 'b' + str(ix))

            # replace all ':' operators with np.mean operations
            bidx = re.findall('[b][0-9]+[:][b][0-9]*[0-9]', op)
            for s in bidx:
                ix0 = int(s.split(':')[0][1:])
                ix1 = int(s.split(':')[1][1:])
                op = op.replace(s, 'np.nanmean( a[..., %d:%d] )' % (ix0, ix1))

            # replace all remaining bands with relevant slices
            bidx = re.findall('[b][0-9]+', op)
            for s in bidx:
                ix = int(s[1:])
                op = op.replace(s, 'a[..., %d]' % ix)

            # finally, remove any $ from constants
            const = re.findall('[$][0-9]+', op)
            for s in const:
                op = op.replace(s, s[1:])

            if print:
                print("Evaluating expression: %s" % op )

            out = self.copy(data=False)
            out.data = eval(op, {'np': np, 'a': self.data})
            if len(out.data.shape) != len(self.data.shape):
                out.data = out.data[..., None] # add extra dim if needed
            out.set_wavelengths(np.arange(out.band_count()))

            return out


    def set_raveled(self, pix, shape=None, onlyFinite = False, strict=True ):
        """
        Fills the image/dataset from a list of pixels of the format returned by get_pixel_list(...). Note that this does not
        copy the list, but simply stores a view of it in this image.

        Args:
            pix (list, ndarray): a list such that pixel[n][band] gives the spectra of the nth pixel.
            shape (tuple): the reshaped data dimensions. Defaults to the shape of the current dataset, except with auto-shape for the last dimension.
            onlyFinite (bool): True if pix contains only pixel values corresponding to non-nan pixels in self.data (as returned by self.X( True ) ).
            strict (bool): True if set_raveled should not change the number of bands in this image. Default is True.
        """
        if shape is None:
            shape = list( self.data.shape )
            shape[-1] = -1

        if strict: # number of bands cannot change
            assert self.data.shape[-1] == pix.shape[-1], \
                "Error: image and pix array have different number of bands. To allow changes to band count please specify strict=False."
            if onlyFinite:
                self.data[ np.isfinite(self.data).all(axis=-1), : ] = pix
            else:
                self.data = pix.reshape(shape)
        else:
            newdata = np.full( self.data.shape[:-1] + (pix.shape[-1],), np.nan, )
            if onlyFinite:
                if onlyFinite:
                    newdata[np.isfinite(self.data).all(axis=-1), :] = pix
                else:
                    newdata = pix.reshape( self.data.shape[:-1] + (pix.shape[-1],) )
            self.data = newdata

    def get_band_index(self, w, **kwds):
        """
        Get the band index that corresponds with the given wavelength or band name.

        Args:
            w (float, int, str): the wavelength, index or band name search for. Note that if w is an integer it is treated
               as a band index and simply returned. If it is a string then the index of the matching band name
               is returned. If it is a wavelength then the closest band to this wavelength is returned.
            **kwds: This function takes one keyword:

                 - thresh = the threshold (in nanometers) within which a band must fall to be valid. Default is
                            hylite.band_select_threshold (which defaults to 10 nm). If a wavelength is passed and a
                            band exists within this distance, then it is returned. Otherwise an error is thrown).

        Returns:
            the matching band index.
        """

        thresh = kwds.get("thresh", hylite.band_select_threshold)

        if np.issubdtype( type(w), np.integer ):  # already a valid band index
            assert -self.band_count() <= w <= self.band_count(), "Error - band index %d is out of range (image has %d bands)." % (w, self.band_count())
            if w < 0: #convert negative indices to positive ones
                return self.band_count() + w
            else:
                return w
        elif isinstance(w, str): # treat w as band name
            assert w in self.get_band_names(), "Error - could not find band with name %s" % w
            return int(list(self.get_band_names()).index(w))
        elif np.issubdtype( type(w), np.floating ): # otherwise treat w as wavelength
            wavelengths = self.get_wavelengths()
            diff = np.abs( wavelengths - w)
            assert np.nanmin(diff) <= thresh, "Error - no bands exist within %d nm of wavelength %f. Try increasing the 'thresh' keyword?" % (thresh, w)
            return int(np.argmin(diff))
        else:
            assert False, "Error - %s is an unknown band descriptor type." % type(w)

    def resample(self, w, agg=True, bw=None, vb=True, partial=False, **kwds):
        """
        Return a copy of this dataset resampled onto the specified wavelength array. Note that this does not
        do any interpolation, but rather selects the nearest band(s) and averages them if need be. Hence it is useful
        for reducing spectral resolution to match e.g. another dataset, but should NOT be used
         to sample bands at higher spectral resolution.

        Args:
            w (ndarray): Either a 1-D numpy array containing the wavelengths to sample onto (see bw for setting the band width), or
               a (n,2) array containing (start, end) wavelength ranges for each band (in this case bw will be ignored).
            agg (bool): True if bands between each entry in w (i.e. w - bw -> w + bw) should be averaged (i.e. spectral binning).
                  Default is True. If False then the closest value will be used (i.e. spectral subsampling).
            bw (float): the width of each band. If None (default) then this is calculated as w[1] - w[0] (i.e. assume regular
                 spacing). This has no effect if agg is False.
            partial (bool): True if partial overlap between the source and target wavelengths is allowed. Non-overlapping areas will be
                      replaced with nan. Default is False.
            vb (bool): True if a progress bar should be created. Default is True.
            **kwds: if provided, the thresh keyword will be passed to get_band_index(...) to control tolerances when selecting the
                closest bands. See documentation for get_band_index(...) for more details.

        Returns:
            a copy of this HyData instance resampled onto the new wavelength array.
        """

        out = self.copy(data=False)  # create output array
        out.data = np.full(self.data.shape[:-1] + (len(w),), np.nan)
        out_wav = []
        w = np.array(w)  # ensure this is an array
        if bw is None and len(w.shape) == 1:
            bw = abs(w[1] - w[0])
        loop = range(len(w))
        if vb:
            loop = tqdm(loop, desc='Resampling bands', leave=False)
        for i in loop:
            idx0 = None
            idx1 = None
            if agg:  # binning
                if len(w.shape) == 1: # W is 1-D
                    out_wav.append(w[i])  # use specified wavelength.
                    try:
                        idx0 = self.get_band_index(w[i] - bw * 0.5, **kwds)
                        idx1 = self.get_band_index(w[i] + bw * 0.5, **kwds)
                    except AssertionError:
                        if partial:
                            pass
                        else:
                            assert False, "Error - source wavelength array (%.1f - %.1f) does not cover target array (%.1f - %.1f)." % (
                                self.get_wavelengths()[0], self.get_wavelengths()[-1], w[0], w[-1])

                else: # W is 1-D, containing (start,end) pairs.
                    out_wav.append((w[i, 0] + w[i, 1]) / 2)  # use midpoint as wavelength
                    try:
                        idx0 = self.get_band_index(w[i,0], **kwds)
                        idx1 = self.get_band_index(w[i,1], **kwds)
                    except AssertionError:
                        if partial:
                            pass
                        else:
                            assert False, "Error - source wavelength array (%.1f - %.1f) does not cover target array (%.1f - %.1f)." % (
                                self.get_wavelengths()[0], self.get_wavelengths()[-1], w[0], w[-1])
                if idx0 is not None and idx1 is not None:
                    out.data[..., i] = np.nanmean(self.data[..., idx0:(idx1 + 1)], axis=-1)
            else:
                assert len(w.shape) == 1, "Error - if agg is False then w must be 1d, not %s" % str(w.shape)
                try:
                    idx = self.get_band_index(w[i], **kwds)
                    out_wav.append(self.get_wavelengths()[idx])  # use real wavelength
                    out.data[..., i] = self.data[..., idx]  # nearest-neighbour resampling
                except AssertionError:
                    if partial:
                        out_wav.append(w[i])  # use passed wavelength
                    else:
                        assert False, "Error - source wavelength array (%.1f - %.1f) does not cover target array (%.1f - %.1f)." % (
                        self.get_wavelengths()[0], self.get_wavelengths()[-1], w[0], w[-1])
        out.set_wavelengths(out_wav)
        return out

    def contiguous_chunks(self, p=75, min_size=0):
        """
        Extract contiguous chunks of spectra, splitting a (1) completely nan bands or (2) large steps in wavelength.

        Args:
            p (int): the percentile used to define a large change in wavelength. Default is 90. A "gap" is considered to be
               a change in wavelength greater than double this percentile.
            min_size (int): the minimum number of bands required to consider a chunk valid. Default is 0 (return all chunks).

        Returns:
            Tuple containing

            chunks (ndarray): copies of the orignal data array that contain continuous spectra. At least one pixel/point
                    in each slice of these bans is guaranteed to be finite.
            wav (ndarray): array containing the wavelengths corresponding to each band of each chunk.
        """

        # find gaps in wavelength and/or completely nan bands and/or data ignore values
        finite = np.isfinite(self.data).any(axis=tuple(range(len(self.data.shape) - 1)))  # False = nans
        finite = finite & (self.data != float(self.header.get("data ignore value", 0))).any(
            axis=tuple(range(len(self.data.shape) - 1)))
        assert len(self.get_wavelengths()) == len(
            finite), "Error - hyperspectral dataset has %d bands but %d wavelengths." % (
        len(finite), len(self.get_wavelengths()))
        x = self.get_wavelengths()[finite]
        dx = np.abs(np.diff(x))
        maxstep = 2. * np.percentile(dx, p)
        if not (dx >= maxstep).any():  # no gaps - just return contiguous block!
            assert len(x) > min_size, "Error - total band count < min_size."
            msk = [self.get_band_index(b) for b in x]
            return [self.data[..., msk]], [x]
        else:
            break_start = list(np.argwhere(dx > maxstep)[:, 0])
            break_end = list((-np.argwhere(np.abs(np.diff(x[::-1])) > maxstep)[:, 0])[::-1])
            break_start.append(-1)  # add end of dataset so we don't miss last chunk
            break_end.append(-1)  # add end of dataset so we don't miss last chunk
            assert len(break_start) == len(
                break_end), r"Error - weird shit is happening? [ useful error messages ftw ¯\_(ツ)_/¯ ]"
            idx0 = 0
            chunks = []
            wav = []
            for i in range(len(break_start)):  # build chunks
                W = x[idx0:break_start[i]]
                msk = [self.get_band_index(b) for b in W]
                if W.shape[-1] > min_size:
                    wav.append(W)
                    chunks.append(self.data[..., msk])
                idx0 = break_end[i]  # skip forwards
            return chunks, wav

    ##################################
    ## Smoothing algorithms
    ###################################
    def smooth_median(self, window=3):
        """
        Applies running median filter on data.

        Args:
            window (int): size of running window, must be int.

        Returns:
            Nothing - overwrites data with smoothed result.
        """

        assert isinstance(window, int), "Error - running window size must be integer."
        if len(self.data.shape) == 3:  # image data
            self.data = ndimage.median_filter(self.data, size=(1, 1, window))
        elif len(self.data.shape) == 2:  # point cloud data
            self.data = ndimage.median_filter(self.data, size=(1, window))
        else:
            assert False, "Error: Run_median does not work on %d-d data." % len(self.data.shape)

    def smooth_savgol(self, window=5, poly=2, chunk=False, **kwds):
        """
        Applies Savitzky-Golay-filter on data.

        Args:
            window (int): size of running window, must be an odd integer.
            poly (int): degree of polynom, must be int.
            chunk (bool): True if the data should be split into chunks (removing e.g. nan bands) before filtering. Use with care!
                   Default is False.
            **kwds: Keywords are passed to scipy.signal.savgol_filter(...).

        Returns:
            A copy of the input dataset with smoothed spectra.
        """

        assert isinstance(window, int), "Error - running window size must be integer."

        # extract contiguous chunks
        if chunk:
            C, w = self.contiguous_chunks(min_size=window)
        else:
            C = [self.data.copy()]
            w = [self.get_wavelengths().copy()]

        # do smoothing
        kwds['window_length'] = window
        kwds['polyorder'] = poly
        kwds['axis'] = -1
        for X in C:
            mask = np.isfinite(X).all(axis=-1)  # remove nans
            X[mask, :] = signal.savgol_filter(X[mask, :], **kwds)

        # return copy
        out = self.copy(data=False)
        if self.is_image():
            out.data = np.dstack(C)
            out.set_wavelengths(np.hstack(w))
        else:
            out.data = np.vstack(C)
            out.set_wavelengths(np.hstack(w))
        return out

    def fill_gaps(self):
        """
        Fill spectral gaps using nearest neighbour interpolation (in the spectral direction). This operation is applied in-place.
        """
        from scipy import ndimage as nd
        for i in range(self.data.shape[0]):
            if self.is_image():
                for j in range(self.data.shape[1]):
                    invalid = np.logical_not(np.isfinite(self.data[i, j, :]))
                    if invalid.any():
                        ind = nd.distance_transform_edt(invalid, return_distances=False, return_indices=True)
                        self.data[i, j, :] = self.data[i, j, tuple(ind)]
            else:
                invalid = np.logical_not(np.isfinite(self.data[i, :]))
                if invalid.any():
                    ind = nd.distance_transform_edt(invalid, return_distances=False, return_indices=True)
                    self.data[i, :] = self.data[i, tuple(ind) ]

    ###################################
    # PLOTTING AND OTHER VISUALISATIONS
    ###################################
    # noinspection PyDefaultArgument
    def plot_spectra(self, ax=None, band_range=None, labels=None, indices=[], colours='blue', **kwds):
        """
        Plots a summary of all the spectra in this dataset.

        Args:
            ax: an axis to plot to. If None (default), a new axis is created.
            band_range (tuple): tuple containing the (min,max) band index (int) or wavelength (float) to plot.
            labels (list): Labels for spectral features such that labels[0] = [feat1,feat2,..] and labels[1] = [name1,name2,...]
                    can be passed. Pass None (default) to disable labels.
            indices (list): specific data point to plot. Should be a list containing index tuples, or an empty list if no pixels
                    should be plotted (Default).
            colours (list,str): a matplotlib colour string or list of colours corresponding to each index spectra. Default is 'blue'.
            **kwds: keywords are passed to plt.plot( ... ), except the following options:

                 - quantiles = True if summary quantiles of all pixels should be plotted. Default is True.
                 - median = True if the median spectra of all pixels should be plotted. Default is True.

        """

        if ax is None:
            fig, ax = plt.subplots(figsize=(18, 6))

        # extract relevant range
        subset = self
        if band_range is not None:
            subset = self.export_bands(band_range)

        # ensure wavelengths are appropriate
        if not subset.has_wavelengths() or len(subset.get_wavelengths()) != subset.band_count():
            subset.set_wavelengths( np.arange( subset.band_count() ) )

        # calculate and plot percentiles
        quantiles = kwds.get("quantiles", True)
        median = kwds.get("median", True)
        if "quantiles" in kwds: del kwds['quantiles']  # remove keyword
        if 'median' in kwds: del kwds['median']  # remove keyword
        C, x = subset.contiguous_chunks()
        for C, x in zip(C, x):
            if quantiles or median:  # calculate percentiles
                percent = np.nanpercentile(C, axis=tuple(range(len(self.data.shape) - 1)),
                                           q=[5, 25, 50, 75, 95])  # calculate percentiles
                q5, q25, q50, q75, q95 = percent

            if median:  # plot median line
                ax.plot(x, q50, color='k', label='median', **kwds)
            if quantiles:  # plot percentile areas
                for lower, upper in zip([q5, q25], [q95, q75]):
                    _y = np.hstack([lower, upper[::-1]])
                    _x = np.hstack([x, x[::-1]])
                    ax.fill(_x[np.isfinite(_y)], _y[np.isfinite(_y)], color='grey', alpha=0.25)

            # plot specific spectra
            if isinstance(indices, tuple): indices = [indices]
            for i,idx in enumerate(indices):
                if isinstance(colours, list):
                    ax.plot(x, C[idx], color=colours[i], **kwds)
                else:
                    ax.plot(x, C[idx], color=colours, **kwds)

        # plot labels
        if not labels is None:  # plot labels?

            # parse string labels
            if isinstance(labels, str):
                if 'silicate' in labels.lower():  # plot common minerals theme
                    labels = ref.Themes.CLAY
                elif 'carbonates' in labels.lower():  # plot carbonate minerals
                    labels = ref.Themes.CARBONATE
                elif 'ree' in labels.lower():  # plot REE features
                    labels = None  # todo - add REE features

            i = 0
            for l in labels:
                # plot label
                if ax.get_xlim()[0] < l.pos < ax.get_xlim()[1]:
                    if isinstance(l, MultiFeature) or isinstance(l, MixedFeature):
                        l.quick_plot(ax=ax, method='fill+line', sublabel=i)
                        i += l.count()
                    else:
                        l.quick_plot(ax=ax, method='fill+line', label=i)
                        i += 1

        # add grid x-ticks
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.grid(which='major', axis='x', alpha=0.75)
        ax.grid(which='minor', axis='x', alpha=0.2)
        ax.set_xlabel("Wavelength (%s)" % self.header.get('wavelength units', 'nm'))
        #ax.set_ylabel("Reflectance") # n.b. not all images contain reflectance data...
        return ax.get_figure(), ax

    ###################################
    ##DATA TRANSFORMS AND COMPRESSION
    ###################################

    def compress(self):
        """
        Convert data array to int16 to save memory.
        """

        #no need to compress...
        if self.data.dtype == np.uint16: return
        if np.nanmin(self.data) < 0:
            print( "Warning - negative values cannot be converted to uint and will be lost. "
                   "Minimum value in dataset is: %s." % np.nanmin(self.data) )

        #map to range 1 - 65535
        sf = 65535 / np.nanmax(self.data)
        self.data *= sf
        self.data[self.data < 0] = 0

        #convert data to uint16
        self.data = self.data.astype(np.uint16)

        #store min/max in header
        self.header["data ignore value"] = str(0)
        self.header['reflectance scale factor'] = sf

    def decompress(self):
        """
        Expand data array to floats to get actual values
        """
        # no need to decompress...
        if self.data.dtype != np.uint16: return

        # get min/max data
        sf = float(self.header.get("reflectance scale factor", 65535))
        nan = float(self.header.get("data ignore value", -1))

        # update header accordingly (to avoid save / load issues in the future!)
        self.header["reflectance scale factor"] = 1.0
        if 'data ignore value' in self.header:
            del self.header['data ignore value']

        # expand data array to float32
        self.data = self.data.astype(np.float32)
        self.data /= sf

        # set nans
        self.set_as_nan(nan)

    def getQuantized(self, n=255, cmeth='KMeans', vthresh=10, smooth=5, subsample=50, mask=None, normalise=False):
        """
        Get a lossy compressed version of this hyperspectral dataset using the "quantisation" method.
        This uses a combination of PCA and K-means to cluster the image into n classes, for which a
        spectral library containing the minimum, median and maximum spectra is derived. This can be used
        to e.g., run computationally expensive operations on relatively few spectra or store lossy compressed
        versions of the hyperspectral data.

        Args:
         n (int): The number of quanta (discrete spectral classes) to use. Default is class 1 to 255, + 0 for background.
         cmeth (str): Method used for clustering. Options are 'Birch' (default), 'KMeans' or 'MiniBatch' (KMeans).
         vthresh (int,float): A float variance threshold (0-1) used to determine the number of PCA bands to use, or an integer
                        to use n-bands. Default is 10 (bands).
         smooth (int): The window width of a Savgol filter used to denoise spectra before fitting the PCA and K-means. Default is 5. Must be an odd integer. Set to 0 to disable.
         subsample (int): The stride used to subsample pixels when fitting the PCA and k-means. Higher values are faster but
                    less accurate. Default is 50.
         mask (ndarray): A binary mask to apply before fitting the PCA and k-means. Masked data will be put in their own
                     class. Data flagged with True in the mask will be retained, while those flagged as False will be removed.
         normalise (bool): If True, PCA components are normalised before k-means clustering. Default is False.
        Returns: An index dataset (HyData) containing the class IDs, and a spectral library containing the minimum, median,
                  and maximum spectra of each class.
        """
        # get data
        if mask is None:
            mask = np.isfinite(self.data).all(axis=-1)  # keep only pixels that are all finite
        else:
            mask = mask & np.isfinite(self.data).all(axis=-1)  # use provided mask & ensure finite
        X = hylite.HyData(self.data[mask, :])
        if smooth > 1:
            X.smooth_savgol(smooth)

        # check subsample makes sense
        if n > (X.data.shape[0] / subsample):
            subsample = 1 # don't do it - this will only happen for small datasets where it's not needed anyway

        # run a PCA
        from hylite.filter import PCA
        nbands = int(self.band_count() / 2)
        if vthresh > 1:
            nbands = int(vthresh)
        pca, _, _ = PCA(X, bands=nbands, step=subsample)

        # normalise and extract bands
        if normalise:
            pca.percent_clip(1, 99, per_band=True, clip=False )
        if vthresh > 1:
            ix = int(vthresh)
        else:
            ix = np.argmax(pca.get_wavelengths() > vthresh)  # get our noise cutoff
        # print(ix)

        # fit k-means, subsampling our pixels for speed
        from sklearn.cluster import KMeans, Birch, MiniBatchKMeans
        if 'birch' in cmeth.lower():
            km = Birch(n_clusters=n).fit(pca.X()[::subsample, :ix])
        elif 'minibatch' in cmeth.lower():
            km = MiniBatchKMeans(n_clusters=n).fit(pca.X()[::subsample, :ix])
        else:
            km = KMeans(n_clusters=n).fit(pca.X()[::subsample, :ix])

        # predict for all valid pixels
        C = km.predict(pca.X()[:, :ix])

        # reshape into original shape
        out = np.zeros(mask.shape + (1,))
        out[mask, 0] = C + 1 # N.B. +1 leaves index 0 free for nans / mask

        # put in relevant type and copy metadata
        index = self.copy(data=False)
        index.data = out.astype(int)
        index.set_band_names(['index'])
        index.set_wavelengths([0])

        # build spectral library
        minHC = []
        medHC = []
        maxHC = []
        counts = []
        for i in range(n+1):
            mask = index.data[..., 0] == i
            X = self.data[mask, :]  # get spectra
            counts.append(X.shape[0])
            if (len(X) > 0):
                mn, md, mx = np.percentile(X, (1, 50, 99), axis=0)  # get min, median and max value in each class
                minHC.append(mn)
                medHC.append(md)
                maxHC.append(mx)
            else:
                minHC.append(np.full( X.shape[-1], np.nan ))
                medHC.append(np.full( X.shape[-1], np.nan ))
                maxHC.append(np.full( X.shape[-1], np.nan ))

        lib = hylite.HyLibrary(np.transpose(np.dstack([minHC, medHC, maxHC]), (0, 2, 1)), wav=self.get_wavelengths())
        if self.has_band_names():
            lib.set_band_names( self.get_band_names() )

        # add class abundances to library header (this can be useful for doing weighted statistics)
        lib.header['counts'] = np.array(counts)

        return index, lib

    @classmethod
    def fromQuanta(cls, index, library ):
        """
        Reconstruct a HyData instance from an index and spectral library, as returned by `HyData.getQuantized(...)`.

        Args:
            index (HyData): A classification index of the same type (e.g., HyImage, HyCloud, etc.) as the desired output.
            library (HyLibrary): A spectral library containing the spectral information associated with each class. Note
                                 that this could be a transformed (e.g., by minimum wavelength mapping) version of the
                                 spectral library created by `getQuantized( ... )`.
        """

        outdata = np.zeros( index.data.shape[:-1] + (library.data.shape[-1],), dtype=library.data.dtype )
        for n in np.unique( index.data[...,0].astype(int) ):
            mask = index.data[...,0] == n
            if n == 0:
                continue # this is the background
            if library.data.shape[1] == 3:
                outdata[mask, : ] = library.data[n,1,:]
            elif library.data.shape[1] == 1:
                outdata[mask, :] = library.data[n, 0, :]
            else:
                assert False, "Error - library should contain either one (median) or three (minimum, median, maximum) " \
                              "samples for each class"

        out = index.copy(data=False)
        out.data = outdata
        out.set_wavelengths( library.get_wavelengths() )
        if (library.has_band_names()):
            out.set_band_names( library.get_band_names() )

        return out

    def normalise(self, minv=None, maxv=None):
        """
        Normalizes individual data points to account for variations in illumination and overall reflectivity. This can be done
        in two ways: if minv and maxv are both none, each pixel vector will be normalized to length 1. Otherwise, if minv
        and maxv are specified, every data point is normalised to the average of the bands between minv and maxv.

        Returns:
            the normalising factor used for each data point.
        """

        # convert to float
        dtype = self.data.dtype
        self.data = self.data.astype(np.float32)

        # normalise pixel vectors
        if minv is None and maxv is None:

            # get valid bands
            valid = []
            for b in range(self.band_count()):
                if np.isfinite(self.get_band(b)).any() and not (self.get_band(b) == 0).all():
                    valid.append(b)

            # easy!
            nf = np.linalg.norm(self.data[..., valid], axis=-1).astype(np.float32)
            self.data /= nf[..., None]

        #normalise to band average
        else:
            # get minimum band to normalize over
            minidx = self.get_band_index( minv )
            maxidx = self.get_band_index( maxv )

            if minidx > maxidx:
                t = minidx
                minidx = maxidx
                maxidx = t

            # normalize
            nf = np.nanmean(self.data[..., minidx:maxidx], axis=-1).astype(np.float32)
            with np.errstate(all='ignore'):  # ignore div 0 errors etc.
                self.data /= nf[..., None]

        #convert back to original datatype
        if np.issubdtype( dtype, np.integer ):
            self.compress() #convert back to integer

        return nf

    def percent_clip(self, minv=2, maxv=98, per_band=False, clip=True):
        """
        Scale self.data such that the specified percentiles become 0 and 1 respectively. Note that this
        normalisation is applied in situ.

        Args:
            minv (int): the lower percentile. Default is 2.
            maxv (int): the upper percentile. Default is 98.
            per_band (bool): apply scaling to bands independently. Default is False.
            clip (bool): True if values < minv or > maxv should be clipped to 0 or 1. Default is True.

        Returns:
            vmin, vmax (float): the percentile clip thresholds used for the normalisation
        """

        # calculate percentile thresholds
        if per_band:
            minv, maxv = np.nanpercentile(self.data, (minv, maxv), tuple(np.arange(len(self.data.shape) - 1)))
        else:
            minv, maxv = np.nanpercentile(self.data, (minv, maxv))

        # ensure dtype is a float
        if np.issubdtype(self.data.dtype, np.integer):
            self.data = self.data.astype(np.float32)

        # apply normalisation
        self.data = (self.data - minv) / (maxv-minv)

        # apply clipping
        if clip:
            self.data = np.clip( self.data, 0, 1 )

        return minv, maxv

    def correct_spectral_shift(self, position):
        """
        Corrects potential spectral sensor shifts by shifting the offset (right) part of the spectrum.

        Args:
            position (float,int): Wavelength or band position of the first offset value - e.g. FENIX: band 714 or wavelength 976., respectively.

        Returns:
            None - changes data in place
        """
        assert isinstance(position, int) or isinstance(position, float), "Error - shift position must be int (band number) or float (wavelength)."
        if isinstance(position, float):
            position = self.get_band_index(position)

        self.data[..., position:] += (self.data[..., (position - 1)] - self.data[..., position])[..., None]


"""
Denoising, feature-extraction and dimensionality reduction filters.
"""

from .dimension_reduction import MNF, PCA, from_loadings
from .segment import *
from .sample import *
from .tpt import *


def boost_saturation( image, bands, flip=True, sat=0.8, val=None, clip=(2,98),per_band=False ):
    """
    Create a saturation boosted composite image.

    Args:
        image (hylite.HyImage): the image containing the data
        bands: the bands to map to r, g, b.
        flip: true if the image should be inverted prior to boosting saturation
              (good for interpreting absorbtion features). Default is True.
        sat: the (constant) saturation value to use. Default is 0.8.
        val: the (constant) brightness value to use. Default is None (do not fix brightness).
        clip: the percentile clip range to use to map to 0 - 1. Default is the 2nd and 98th percentile (2,98).
        per_band: True if this clipping should be done per band or over the whole RGB composite. Default is False
    Returns:
        a HyImage instance containing the saturation enhanced result
    """

    from matplotlib.colors import rgb_to_hsv, hsv_to_rgb

    # subset
    rgb = image.export_bands(bands)

    # clip
    _ = rgb.percent_clip(clip[0], clip[1], per_band=per_band)

    # invert?
    if flip:
        rgb.data = 1 - rgb.data

    hsv = rgb_to_hsv(rgb.data) # map to hsv space

    if sat is not None: # boost sat
        hsv[...,1] = sat
    if val is not None: # boost brightness
        hsv[..., 2] = val

    rgb.data = hsv_to_rgb(hsv) # map back to RGB space

    return rgb


"""
Contains measured reflectance spectra for Spectralon targets. Modify these or add new .txt files to change/add reference
Spectralon targets. For example *MySpeccySpectra.txt* would become accessible as
*hylite.reference.spectra.MySpeccySpectra*.
"""


"""
Save and load viewing positions associated with data acquisition or rendering 3D point cloud data.
"""

import numpy as np

from hylite.io import makeDirs
from hylite.project import Camera, Pushbroom


def saveCameraTXT( path, camera ):
    """
    Saves a Camera view to a simple text format. The camera can either be an instance of Camera or
    Pushbroom.

    Args:
        path: the path to save the file.
        camera: a camera object containing the data to save.
    """

    # make directories if need be
    makeDirs( path )

    with open(path,'w') as f:

        if isinstance( camera, Camera ):
            f.write("Camera # camera type\n")
            f.write('%.3f,%.3f,%.3f #camera position\n' % (camera.pos[0],camera.pos[1],camera.pos[2]))
            f.write('%.3f,%.3f,%.3f #camera ori\n' % (camera.ori[0], camera.ori[1], camera.ori[2]))
            f.write('%s #projection type\n' % camera.proj)
            f.write('%.3f #vertical field of view (deg)\n' % camera.fov)
            f.write('%d,%d #dims\n' % (camera.dims[0],camera.dims[1]))
            if 'pano' in camera.proj.lower():
                f.write('%.3f # angular pitch (x)\n' % camera.step)
        elif isinstance( camera, Pushbroom ):
            f.write("Pushbroom # camera type\n")
            f.write('%.3f,%.3f # xfov, yfov\n' % (camera.xfov, camera.lfov))
            f.write('%d,%d # xdim, ydim\n' % (camera.dims[0], camera.dims[1]))
            for i in range(3): # write position arrays
                f.write("%s\n" % ','.join(["%.6f" % v for v in camera.cp[:,i]]))
            for i in range(3):  # write orientation arrays
                f.write("%s\n" % ','.join(["%.6f" % v for v in camera.co[:, i]]))
        else:
            assert False, '%s is not a Camera or Pushbroom object' % type(camera)


def loadCameraTXT( path ):
    """
    Loads a Camera view from a simple text format.

    Returns:
        a Camera object containing the camera properties.
    """

    with open(path, 'r') as f:
        lines = f.readlines()

        #trim comments
        lines = [l.split('#')[0].strip() for l in lines]
        if 'camera' in lines[0].lower():
            pos = np.fromstring( lines[1], sep=',')
            ori = np.fromstring( lines[2], sep=',')
            proj = lines[3]
            fov = float(lines[4])
            dims = tuple( np.fromstring( lines[5], sep=',').astype(int))
            step  = None
            if len(lines) > 6:
                step = float( lines[6] )
            return Camera(pos, ori, proj, fov, dims, step)
        elif 'pushbroom' in lines[0].lower():

            xfov, lfov = np.fromstring( lines[1], sep=',')
            dims = tuple( np.fromstring( lines[2], sep=',').astype(int))
            cp = np.array([ np.fromstring(lines[3+i], sep=',') for i in range(3)]).T
            co = np.array([ np.fromstring(lines[6+i], sep=',') for i in range(3)]).T

            return Pushbroom( cp, co,  xfov, lfov, dims )


"""
Read common image formats, including ENVI format hyperspectral data.
"""

import sys, os
import numpy as np
import spectral
from hylite.hyimage import HyImage, HyData
from .headers import matchHeader, makeDirs, loadHeader, saveHeader

# spectral python throws depreciation warnings - ignore these!
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

def loadWithGDAL(path, dtype=np.float32, mask_zero = True):
    """
    Load an image using gdal.

    Args:
        path: file path to the image to load
        mask_zero: True if zero values should be masked (replaced with nan). Default is true.
    Returns:
        a hyImage object
    """

    # find GDAL
    try:
        import osgeo.gdal as gdal
        gdal.PushErrorHandler('CPLQuietErrorHandler') # ignore GDAL warnings
    except:
        assert False, "Error - please install GDAL before using loadWithGDAL(...)"

    #parse file format
    _, ext = os.path.splitext(path)
    # load envi file or its variants
    if len(ext) == 0 or 'hdr' in ext.lower() or \
            'dat' in ext.lower() or \
            'img' in ext.lower() or \
            'lib' in ext.lower():
        header, image = matchHeader(path)
    elif 'tif' in ext.lower() or 'png' in ext.lower() or 'jpg' in ext.lower(): #standard image formats
        image = path
        header = None
    else:
        print( 'Warning - %s is an unknown/unsupported file format. Trying to load anyway....' % ext)
        header, image = matchHeader(path)
    # load header
    if not header is None:
        header = loadHeader(header)

    #load image
    assert os.path.exists(image), "Error - %s does not exist." % image
    try:
        raster = gdal.Open(image)  # open image
        data = raster.ReadAsArray().T
    except:
        assert False, "Error - %s could not be read by GDAL." % image

    #create image object
    assert data is not None, "Error - GDAL could not retrieve valid image data from %s" % path
    pj = raster.GetProjection()
    gt = raster.GetGeoTransform()
    img = HyImage(data, projection=pj, affine=gt, header=header, dtype=dtype)

    if mask_zero and (img.dtype == np.float32 or img.dtype == np.float64):
            img.data[img.data == 0] = np.nan #note to self: np.nan is float...

    return img

def loadWithSPy( path, dtype=np.float32, mask_zero = True):
    """
    Load an image using spectral python. This works for most envi images, but doesn not load
    georeferencing information (in which case loadWithGDAL(...) should be used).

    Args:
        path: file path to the image to load
        mask_zero: True if zero values should be masked (replaced with nan). Default is true.
    Returns:
        a hyImage object
    """

    assert os.path.exists(path), "Error - %s does not exist." % path

    # parse file format
    _, ext = os.path.splitext(path)
    if len(ext) == 0 or 'hdr' in ext.lower() or \
            'dat' in ext.lower() or \
            'img' in ext.lower() or \
            'lib' in ext.lower():
        header, image = matchHeader(path)

        # load image with SPy
        assert os.path.exists(image), "Error - %s does not exist." % image
        try: # try loading envi file first
            img = spectral.envi.open(header, image) # this must be an envi file
        except:
            img = spectral.open_image(header) # load unknown image type

        data = np.transpose( np.array(img.load()), (1,0,2) )
        if (data.dtype == np.float32) or (data.dtype == np.float64):
            # Spy still divides float values by scale factor. We don't want it to, so undo this.
            data *= img.scale_factor

        # load header
        if not header is None:
            header = loadHeader(header)
    elif 'tif' in ext.lower() or 'png' in ext.lower() or 'jpg' in ext.lower():  # standard image formats
        # load with matplotlib
        import matplotlib.image as mpimg
        data = mpimg.imread(path)
        header = None
    else:
        print('Warning - %s is an unknown/unsupported file format. Trying to load anyway...'%ext)
        #assert False, "Error - %s is an unknown/unsupported file format." % ext

    # create image object
    assert data is not None, "Error - GDAL could not retrieve valid image data from %s" % path
    img = HyImage(data, projection=None, affine=None, header=header, dtype=dtype)

    # spectral python automatically applies reflectance scale factor, so we must set this to 1.0 to avoid future nightmares...
    img.header['reflectance scale factor'] = 1.0

    if mask_zero and img.dtype == float:
        img.data[img.data == 0] = np.nan  # note to self: np.nan is float...

    return img

def loadSubset( path, *, bands=None, pixels=None, dtype=np.float32):
    """
    Load either specific bands (bands!=None) or pixels (pixels != None) from an ENVI file using spy to facilitate e.g. out-of-core
    processing routines.

    Args:
        path: a path to the hyperspectral image to read.
        bands: a list of hyperspectral band indices or wavelengths to extract, or None.
        pixels: a list of [(x1,y1),(x2,y2)] pixels to extract spectra for, or None. Either bands or pixels must be defined (but not both).
        dtype: the output data type. Default is float32.
    """
    assert os.path.exists(path), "Error - %s does not exist." % path
    assert (pixels is not None) or (bands is not None), "Error - either pixels OR bands must be specified"
    assert not ((pixels is not None) and (bands is not None)), "Error - pixels AND bands cannot both be specified"

    # parse file format
    _, ext = os.path.splitext(path)
    if len(ext) == 0 or 'hdr' in ext.lower() or \
            'dat' in ext.lower() or \
            'img' in ext.lower() or \
            'lib' in ext.lower():
        header, image = matchHeader(path)

        # load header and convert bands to band indices
        imageheader = loadHeader(header)
        if bands is not None:
            bands = [ HyImage( np.zeros((3,3,imageheader.band_count())),
                        header=imageheader, wav=imageheader.get_wavelengths() ).get_band_index(b) for b in bands ]

        # load image with SPy
        assert os.path.exists(image), "Error - %s does not exist." % image
        try:  # try loading envi file first
            img = spectral.envi.open(header, image)  # this must be an envi file
        except:
            img = spectral.open_image(header)  # load unknown image type

        if bands is not None:  # get bands and put in HyImage
            data = np.dstack( [ img.read_band( b ).T for b in bands ] )
            out = HyImage( data, projection=None, affine=None, header=imageheader, dtype=dtype)
            out.set_wavelengths( imageheader.get_wavelengths()[bands] )
            if out.has_band_names():
                out.set_band_names( imageheader.get_band_names()[bands])
        if pixels is not None:  # get pixels and put in HyCloud
            data = np.array( [ img.read_pixel( *p[::-1] ) for p in pixels ] )
            out = HyData( data )
            out.header=imageheader
        return out





# noinspection PyUnusedLocal
def saveWithGDAL(path, image, writeHeader=True, interleave='BSQ'):
    """
    Write this image to a file.

    Args:
        path: the path to save to.
        image: the image to write.
        writeHeader: true if a .hdr file will be written. Default is true.
        interleave: data interleaving for ENVI files. Default is 'BSQ', other options are 'BIL' and 'BIP'.
    """

    # find GDAL
    try:
        import osgeo.gdal as gdal
        gdal.PushErrorHandler('CPLQuietErrorHandler') # ignore GDAL warnings
    except:
        assert False, "Error - please install GDAL before using saveWithGDAL(...)"

    # make directories if need be
    makeDirs( path )

    path, ext = os.path.splitext(path)

    if "hdr" in str.lower(ext):
        ext = ".dat"

    #get image driver
    driver = 'ENVI'
    if '.tif' in str.lower(ext):
        driver = 'GTiff'

    #todo - add support for png and jpg??

    #set byte order
    if 'little' in sys.byteorder:
        image.header['byte order'] = 0
    else:
        image.header['byte order'] = 1

    #parse data type from image array
    data = image.data
    dtype = gdal.GDT_Float32
    image.header["data type"] = 4
    image.header["interleave"] = str.lower(interleave)
    if image.data.dtype == np.intc or image.data.dtype == np.int32:
        dtype = gdal.GDT_Int32
        image.header["data type"] = 3
    if image.data.dtype == np.int16:
        dtype = gdal.GDT_Int16
        image.header["data type"] = 2
    if image.data.dtype == np.uint8:
        data = np.array(image.data, np.dtype('b'))
        dtype = gdal.GDT_Byte
        image.header["data type"] = 1
    if image.data.dtype == np.uint or image.data.dtype == np.uint32:
        dtype = gdal.GDT_UInt32
        image.header["data type"] = 13
    if image.data.dtype == np.uint16:
        dtype = gdal.GDT_UInt16
        image.header["data type"] = 12

    #write
    if driver == 'GTiff':
        output = gdal.GetDriverByName(driver).Create( path + ext, image.xdim(), image.ydim(), image.band_count(), dtype)
    else:
        output = gdal.GetDriverByName(driver).Create( path + ext, image.xdim(), image.ydim(), image.band_count(), dtype, ['INTERLEAVE=%s'%interleave] )

    #write bands
    for i in range(image.band_count()):
         rb = output.GetRasterBand(i+1)
         rb.WriteArray(data[:, :, i].T)
         rb = None #close band
    output = None #close file

    if writeHeader and not image.header is None: #write .hdr file
        image.push_to_header()
        saveHeader(path + ".hdr", image.header)

    # save geotransform/project information
    output = gdal.Open(path + ext, gdal.GA_Update)
    output.SetGeoTransform(image.affine)
    if not image.projection is None:
        output.SetProjection(image.projection.ExportToPrettyWkt())
    output = None  # close file

def saveWithSPy( path, image, writeHeader=True, interleave='BSQ'):
    # make directories if need be
    makeDirs(path)

    path, ext = os.path.splitext(path)

    # make sure extension is proper

    if "hdr" in str.lower(ext) or ext == '':
        ext = ".dat"

    # set byte order
    if 'little' in sys.byteorder:
        image.header['byte order'] = 0
        byteorder = 0
    else:
        image.header['byte order'] = 1
        byteorder = 1

    image.push_to_header()
    spectral.envi.save_image( path + ".hdr", np.transpose(image.data,(1,0,2)),
                                    dtype=image.data.dtype, force=True,
                                    ext=ext, byteorder=byteorder, metadata=image.header)
